(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker,                       *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

(* Translate Hoisted, Closure Converted, Anormalized Ycalculus into Talx86 *)

open Utilities
open Ident
open Variable
open Tal
open Talcon
open Yprim
open Ycon
open Ycalculus
open Format
open Talpp

let wordsize = 4                            (* machine word size *)
let floatsize = 8                           (* machine float size *)
let byte4 = Cprim(PCbytes Byte4)            (* 4 byte/integer type *)
let ynorm = ynormalize_subst                (* ycalculus normalization *)
 
(* Code block with instruction list rather than instruction vector *)
type code_blockl = label * con * instruction list

(* Error handling *)
exception Cogen_con of string
exception Cogen_con_unimplemented of string

let die s = (print_string "Cogen_con error: ";print_string s;print_string"\n";
             raise (Cogen_con s))

let debug s = (open_vbox 0; print_string "\n"; print_string s; 
               print_string "\n"; 
               print_flush ())

let handle e message =
  try e () with
    x -> debug message; raise x


let dumpye e = (open_vbox 0; print_newline (); 
               Ypp.print_term e; print_newline (); close_box ())

let talppopt = 
  { Talpp.style = Talpp.MASM; 
    Talpp.kinds = true; 
    Talpp.cons = true}

let dumptc c = Talpp.print_con Format.std_formatter talppopt c


(* Get the size of a normalized ycalculus constructor.  
   All but YCreal is assumed to be wordsize. 
   If we add the Tal kind structure to the ycalculus, then we
   will use delta.
*)

let ysizeof delta yc =
  (match yc with
    YCconst(YCreal) ->  floatsize
  | _ -> wordsize)

(* Get stack to first stack variable *)
let get_sigma c =
  let rec aux c = 
    match c with
      Cvar v -> (Cempty, false)
    | Cempty -> (Cempty,true)
    | Ccons (f,c) -> let (c', continue) = aux c in (Ccons (f, c'), continue)
    | Cappend (c1, c2) ->
      	let (c1', continue) = aux c1 in
      	if continue then
	  let (c2', continue) = aux c2 in
	  (Cappend (c1',c2'), continue)
      	else
	  (c1', continue)
  | _ -> die "get_sigma: stack is ill-formed"
  in
  fst (aux c)

(* Code Generation Assumptions:
   1 All bound identifiers (value and type identifiers) are unique
   2 All functions are closed.  The only free identifiers are labels
      bound in the heap.
   3 All code has been hoisted to the top level. ie: every module
      is a series of lettype definitions followed by a letrec definition
      containing only code followed by an expression containing no
      abstractions.
   4 All code is in anormal form as defined in mil/yanormal with the
     one exception that we allow packs of tuples as right-hand-sides
     as well.
*)
      
(* Code Generation:

   Does not assume constructors are normalized.  Normalizes constructors
   and substitutes all let-bound types during the translation.

   Calling convention:
   Push all arguments on the stack in right to left order.  Then
   push the return type on the stack and jump to the function(ie: use
   the call instruction).  
   esp holds the stack pointer.
   ebp holds a pointer to the current exception handler (allocated on
   stack).
   The result of the function is placed in register eax.
   Exception handlers expect exceptions to be raised with the exception
   in register eax.

   These conventions are revealed by the type translation for functions.
   |A(a1,...,am).(t1,...,tn) -> t| is (fields are initialized by default and 
   sptr == PCstackptr):

   A(a1,...,am,p1,p2).{
     esp: sptr(tra :: |t1| :: ... :: |tn| :: p1 @ texn),
     ebp: sptr(texn)
   }
   
   where
    tra = A().{esp: sptr(p1@texn), 
               ebp: sptr(texn), 
               eax: |t|}                       (* return address type *)
    texn = A().{esp: sptr(p2),eax: exn} :: p2  (* exception frame type *)
*)

(* Translate kinds.  YKtype -> K4byte.  In other words, I am
   assuming the Ycalculus is fully boxed for now. *)
let rec cogen_kind k =
  match k with
    YKtype ->  K4byte
  | YKarrow (k1, k2) -> Karrow (cogen_kind k1, cogen_kind k2)
  | YKprod ks -> Kprod (List.map cogen_kind ks)
  | YKsum ks -> Ksum (List.map cogen_kind ks)

(* Translate cons *)
(* empty register set *)
let empty_regs = Dict.empty (compare)

(* cons hd onto tail *)
let cons hd tl =
  Ccons (hd, tl)

let make_stack args = List.fold_right cons args Cempty

(* abstract variables of kind stack *)
let stacklams vs c =
  List.fold_right (fun v c -> Clam (v, Kstack, c)) vs c

(* kind abbreviations *)
let ka k1 k2 = Karrow (k1, k2)
let ks = Kstack
let k4 = K4byte

(* Global Type Variables *)
let gexn_stack_var = pvar "texn"
let gretaddr_var = pvar "tret"
let gone_arg_var = pvar "t1"  (* Non-polymorphic One K4byte arg functions *)
let gtwo_arg_var = pvar "t2"  (* Non-poly Two K4byte arg functions *)
let gmulti_arg_var = pvar "tmf" (* Non-poly Multi arg function *)
(*
let gone_arg_poly_var = pvar "t1p" (* Polymorphic One K4byte arg function *)
let gtwo_arg_poly_var = pvar "t2p" (* Polymorphic Two K4byte arg function *)
let ggeneral_fun_var = pvar "tfun" (* General functions *)
*)
(* Abbreviations of Common Types *)
let exn_stack_abbrev rho2 =
 Capp (Cvar gexn_stack_var, Cvar rho2)  

let retaddr_abbrev rho1 rho2 c =
  Capp (Capp (Capp (Cvar gretaddr_var, Cvar rho1), Cvar rho2), c)

let one_arg_abbrev arg1 c = 
  Capp (Capp (Cvar gone_arg_var, arg1), c)

let two_arg_abbrev arg1 arg2 c =
  Capp (Capp (Capp (Cvar gtwo_arg_var, arg1), arg2), c)

(* args is a stack of args *)
let multi_arg_abbrev args c =
  Capp (Capp (Cvar gmulti_arg_var, make_stack args), c)

(*
let one_arg_poly_abbrev vks arg1 c =
  let rho1 = newvar () in
  let rho2 = newvar () in
  stacklams [rho1;rho2]
    (List.fold_right (fun (v,k) c -> Cforall (v,k,c)) vks
       (Capp (Capp (Capp (Capp (Cvar gone_arg_poly_var, 
	  Cvar rho1), Cvar rho2), arg1), c)))

let two_arg_poly_abbrev vks arg1 arg2 c =
  let rho1 = newvar () in
  let rho2 = newvar () in
  stacklams [rho1;rho2]
    (List.fold_right (fun (v,k) c -> Cforall (v,k,c)) vks
       (Capp (Capp (Capp (Capp (Capp (Cvar gtwo_arg_var, 
	   Cvar rho1), Cvar rho2), arg1), arg2), c)))

let general_fun_abbrev2 rho1 rho2 vks args c =
  stacklams [rho1;rho2]
    (List.fold_right (fun (v,k) c -> Cforall (v,k,c)) vks
       (Capp (Capp (Capp (Capp (Cvar ggeneral_fun_var, 
	  Cvar rho1), Cvar rho2), make_stack args), c)))

let general_fun_abbrev vks args c =
  let rho1 = newvar () in
  let rho2 = newvar () in
  general_fun_abbrev2 rho1 rho2 vks args c
*)
(* Meta-level functions for type generation *)
(* Generate stack with exn_handler on top of stack var rho2 *)
let cogen_exn_stack rho2 =
  (* Current exception handler type *)
  let handler_code = Ccode 
      (Dict.insert 
	 (Dict.insert empty_regs 
	    Esp (Capp (Cprim PCstackptr, Cvar rho2)))
	 Eax (Cprim PCexn))
  in
  (* Type of exception stack *)
  cons handler_code (Cvar rho2)

(* Generate code for the return address type *)
let cogen_retaddrc rho1 exn_stack c =
  Ccode
    (Dict.insert
       (Dict.insert
	  (Dict.insert empty_regs 
	     Esp (Capp (Cprim PCstackptr, Cappend (Cvar rho1, exn_stack))))
	  Ebp (Capp (Cprim PCstackptr, exn_stack)))
       Eax c)

(* Generate frame type *)
let cogen_frame_type retaddrc cs =
  cons retaddrc 
    (List.fold_right (fun c cs -> cons c cs)
       cs Cempty)

(* Generate register state for function *)
let cogen_reg_state_type rho1 rho2 frame_type exn_stack =
  let full_stack = Cappend (frame_type, Cappend (Cvar rho1, exn_stack)) in 
  Ccode (Dict.insert 
	   (Dict.insert empty_regs 
	      Esp (Capp (Cprim PCstackptr, full_stack)))
	   Ebp (Capp (Cprim PCstackptr, exn_stack)))

(* Generate function type *)
let cogen_function_type rho1 rho2 vks frame_type exn_stack =
  let regs = cogen_reg_state_type rho1 rho2 frame_type exn_stack in
  List.fold_right (fun (v,k) body -> 
    Cforall(v, k, body)) vks 
    (Cforall (rho1, Kstack, Cforall (rho2, Kstack, regs)))

(* Definitions of abstracted types *)
let exn_stack_kind = ka ks ks 
let exn_stack_def =
  let rho1 = newvar () in
  Clam (rho1, Kstack, cogen_exn_stack rho1)

let retaddr_kind = ka ks (ka ks (ka k4 k4))
let retaddr_def =
  let rho1 = newvar () in
  let rho2 = newvar () in
  let retc = newvar () in
  stacklams [rho1;rho2] 
    (Clam (retc, K4byte, 
	   cogen_retaddrc rho1 (exn_stack_abbrev rho2) (Cvar retc)))

let one_arg_kind = ka k4 (ka k4 k4)
let one_arg_def =
  let rho1 = newvar () in
  let rho2 = newvar () in
  let arg = newvar () in
  let res = newvar () in
  Clam (arg, K4byte, 
	Clam (res, K4byte, 
	      cogen_function_type rho1 rho2 [] 
		(cogen_frame_type 
		   (retaddr_abbrev rho1 rho2 (Cvar res))
		   [Cvar arg]) (exn_stack_abbrev rho2)))

let two_arg_kind = ka k4 (ka k4 (ka k4 k4))
let two_arg_def =
  let rho1 = newvar () in
  let rho2 = newvar () in
  let arg1 = newvar () in  
  let arg2 = newvar () in
  let res = newvar () in
  Clam (arg1, K4byte,
    Clam (arg2, K4byte, 
	Clam (res, K4byte, 
	      cogen_function_type rho1 rho2 [] 
		(cogen_frame_type 
		   (retaddr_abbrev rho1 rho2 (Cvar res))
		   [Cvar arg1; Cvar arg2]) (exn_stack_abbrev rho2))))

let multi_arg_kind = ka ks (ka k4 k4)
let multi_arg_def =
  let rho1 = newvar () in
  let rho2 = newvar () in
  let arg = newvar () in
  let res = newvar () in
  Clam (arg, Kstack, 
	Clam (res, K4byte, 
	      cogen_function_type rho1 rho2 [] 
		(cons 
		   (retaddr_abbrev rho1 rho2 (Cvar res))
		   (Cvar arg)) (exn_stack_abbrev rho2)))
(* 
let one_arg_poly_kind = ka ks (ka ks (ka k4 (ka k4 k4))) 
let one_arg_poly_def =
  let rho1 = newvar () in
  let rho2 = newvar () in
  let arg = newvar () in
  let res = newvar () in
  stacklams [rho1;rho2]
    (Clam (arg, K4byte, 
	Clam (res, K4byte, 
	      cogen_reg_state_type rho1 rho2
		(cogen_frame_type 
		   (retaddr_abbrev rho1 rho2 (Cvar res))
		   [Cvar arg]) (exn_stack_abbrev rho2))))

let two_arg_poly_kind = ka ks (ka ks (ka k4 (ka k4 (ka k4 k4)))) 
let two_arg_poly_def =
  let rho1 = newvar () in
  let rho2 = newvar () in
  let arg1 = newvar () in
  let arg2 = newvar () in
  let res = newvar () in
  stacklams [rho1;rho2]
    (Clam (arg1, K4byte,
       Clam (arg2, K4byte,
	 Clam (res, K4byte, 
	    cogen_reg_state_type rho1 rho2
		(cogen_frame_type 
		   (retaddr_abbrev rho1 rho2 (Cvar res))
		   [Cvar arg1; Cvar arg2]) (exn_stack_abbrev rho2)))))

let general_fun_kind = ka ks (ka ks (ka ks (ka k4 k4)))
let general_fun_def =
  let rho1 = newvar () in
  let rho2 = newvar () in
  let arg = newvar () in
  let res = newvar () in
  stacklams [rho1;rho2;arg] 
	(Clam (res, K4byte, 
	      cogen_reg_state_type rho1 rho2
		 (cons (retaddr_abbrev rho1 rho2 (Cvar res)) (Cvar arg)) 
		 (exn_stack_abbrev rho2)))
*)
let toplevel_typedefs () =
[|gexn_stack_var, exn_stack_def; 
  gretaddr_var, retaddr_def;
  gone_arg_var, one_arg_def;
  gtwo_arg_var, two_arg_def;
  gmulti_arg_var, multi_arg_def
(*
  gone_arg_poly_var, one_arg_poly_kind, one_arg_poly_def;
  gtwo_arg_poly_var, two_arg_poly_kind, two_arg_poly_def;
  ggeneral_fun_var, general_fun_kind, general_fun_def
*)
|]

let wrap_forall vks c =
  List.fold_right (fun (v,k) c -> Cforall (v, cogen_kind k, c)) vks c

(* Translate Cons *) 
let rec cogen_con c =
  let cogen_cconst yconst =
    (match yconst with
      YCint -> byte4
    | YCstring -> Cprim (PCbytevector Byte1) (* Strings are byte1 vectors *)
    | YCreal -> raise (Cogen_con_unimplemented "YCreal -> boxed real?")
    | YCexn -> Cprim PCexn
    | YCexnname -> Cprim PCexnname
    (* Refs become Read/Write tuples.  Note, I'm boxing everything. *)
    | YCref -> 
	let v = newvar () in 
	Clam (v, K4byte, Cprod [(Cvar v, ReadWrite, Init)])
    | YCarray -> Cprim PCarray
    | YCvector -> Cprim PCvector)
  in
  match c with
    YCvar v -> Cvar v
  | YCconst yconst -> cogen_cconst yconst
  | YClam (v,k,c) -> Clam(v, cogen_kind k, cogen_con c)
  | YCapp (c1, c2) -> Capp (cogen_con c1, cogen_con c2)
  | YCtuple cs -> Ctuple (cogen_cons cs)
  | YCproj (i,c) -> Cproj(i,cogen_con c)
  | YCtag (i,c,ks) -> Cin (i, cogen_con c, 
			   Kprod (List.map cogen_kind ks))
  | YCdecide (c,vkcs,k) -> 
      Ccase (cogen_con c, 
	     List.map (fun (v,k,c) -> 
	       (v, cogen_kind k, cogen_con c)) vkcs)
  | YCrec (k,c) -> raise (Cogen_con_unimplemented "cogen_con: YCrec")
  | YCmu (v,c) -> Crec ([(v,K4byte,cogen_con c)], v)
  | YCabs (vks,[arg],c) -> 
      wrap_forall vks (one_arg_abbrev (cogen_con arg) (cogen_con c))
  | YCabs (vks,[arg1;arg2],c) -> 
      wrap_forall vks 
	(two_arg_abbrev (cogen_con arg1) (cogen_con arg2) (cogen_con c))
  | YCabs (vks,cs,c) ->
      wrap_forall vks (multi_arg_abbrev (cogen_cons cs) (cogen_con c))
      (*
      general_fun_abbrev 
	(List.map (fun (v,k) -> (v, cogen_kind k)) vks)
	(cogen_cons cs) (cogen_con c)
      *)
      (*
      let rho1 = newvar () in
      let rho2 = newvar () in
      let exn_stack = cogen_exn_stack rho2 in
      let tra = cogen_retaddrc rho1 exn_stack (cogen_con c) in
      let frame_type = cogen_frame_type tra (cogen_cons cs) in
      cogen_function_type rho1 rho2 
	(List.map (fun (v,k) -> (v,cogen_kind k)) vks) 
	frame_type exn_stack
      *)
  | YCexist (v,k,c) -> Cexist (v, cogen_kind k, cogen_con c)
  | YCprod cs -> 
      (* Unit/Sum [] => Singleton(0) *)
      (match cs with
	[] -> Cprim (PCtag 0) (* Singleton(0) *)
      |	_ ->  Cprod (List.map (fun c -> (cogen_con c, Read, Init)) cs))
  | YCsum css ->
      (* I don't assume that the constructors are sorted.  tag_split 
	 separates the constant constructors from the value-carrying
	 constructors of the sum and translates the types of the
	 constructors for the tvs (tuples or variants) *)
      let rec tag_split css tagnum tvnum =
	match css with
	  [] -> ([],[])
	| []::rest ->
	    let (tags,tvs) = tag_split rest (tagnum + 1) tvnum in
	    (tagnum::tags, tvs)
	| cs::rest ->
	    let (tags, tvs) = tag_split rest tagnum (tvnum + 1) in
	    (tags, 
	     (tvnum, List.map (fun c -> (cogen_con c, Read, Init)) cs)::tvs)
      in
      (* Here we create the sum, changing a single Variant to a Tuple *)
      let mk_sum (tags, vts) =
	{sum_tags = tags;
	 sum_vt = 
	  (match vts with
	    [] -> None
	  | [(_,cs)] -> Some (Tuple cs)
	  | vts -> Some (Variants vts))}
      in
      Csum (mk_sum (tag_split css 0 0))
  | YCsingle i -> Cprim (PCtag i)
  | YCunit -> Cprim (PCtag 0) (* Singleton(0) *)
  | YCvoid -> Csum {sum_tags = []; sum_vt = None}
and cogen_cons cs = List.map cogen_con cs

(* Get sinfo (stack information): 
     (retaddrc, ret_off, ssz, sigma, rho1, rho2) from 
   a label_state type and the constructors of the ycalculus arguments to 
   the function. 
*)

let tal_function_type delta c =
  match c with
    YCabs (vks, cs, c) ->
      let rho1 = newvar () in
      let rho2 = newvar () in
      let taldelta = (rho1, rho2) in
      let ssz = List.fold_right (fun c ssz -> 
	ysizeof delta c + ssz) cs wordsize
      in
      let exn_stack = exn_stack_abbrev rho2 in
      let retaddrc = retaddr_abbrev rho1 rho2 (cogen_con c) in
      let talcs = cogen_cons cs in
      let sigma = cogen_frame_type retaddrc talcs in
      let tc = cogen_function_type rho1 rho2 
	  (List.map (fun (v,k) -> (v, cogen_kind k)) vks) sigma exn_stack 
      in
      (tc, taldelta, retaddrc)
  | _ -> die "tal_context: constructor not function type"

