(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker,                       *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

open Utilities
open Ident
open Variable
open Tal
open Talcon
open Yprim
open Ycon
open Ycalculus
open Format
open Talpp
open Cogen_error
open Cogen_con
open Cogen_util
open Cogen_val

(* rhs_finish: 
     Bind_id: the result of the ycalculus computation is bound to 
              ident, put the result anywhere, updating layout.
     Bind_loc: the result of the computation is constrained to loc.
              Do not update layout.
*)
type rhs_finish =
    Bind_id of ident
  | Bind_loc of reg

(* Compute primitive operation.  Place result in register rdest and
   continue.  I don't check for overflow in any arithmetic operations.
*)
let cogen_prim (p,cs,es)
    (((delta, gamma, kappa) as ctxt,
      ((ra, regs, frame, ret, deadvars) as layout),
      ((rho1, rho2) as taldelta)) as env) 
    dest cont =
  (* Compute the binary arithmetic op given by ab on es and continue with 
     cont.
  *)
  let arithbin ab es =
    (* return true if arithbin ab is symmetric modulo putting result in dest
       register. 
    *)
    let symm_ab ab =
      match ab with
    	Adc -> true
      | Add -> true
      | And -> true
      | Imul2 -> true
      | Or -> true
      | Sbb -> false
      | Sub -> false
      | Xor -> true
    in
    match es with
      [e1; e2] -> 
	begin
	match dest with
	  Bind_id x ->
            let ((raw1,clist1) as e1', _) = cogen_val env e1 in
	    let ((raw2,clist2) as e2',_) = cogen_val env e2 in
	    if clist1 != [] or clist2 != [] then 
	      die "arithbin: coerced args";
	    if is_dead deadvars x e1 & valid_binops raw1 raw2 then
	      singlei (ArithBin (ab, raw1, raw2)) 
		(YCconst YCint, replace_id layout x e1) cont
	    else if symm_ab ab & valid_binops raw1 raw2 &
	            is_dead deadvars x e2 
	    then
	      singlei (ArithBin (ab, raw2, raw1))
		(YCconst YCint, replace_id layout x e2) cont
	    else 
	      begin
	      match raw1, raw2 with
		Reg r1, _ ->
		  let (spill_i, layout') = spill_reg layout r1 in
		  let (il, cbs) = cont (YCconst YCint, add_id layout' x (Rloc r1))
		  in
		  (spill_i @
		   (ArithBin (ab, Reg r1, raw2) ::
		    il), cbs)
	      |	_, Reg r2 ->
		  if symm_ab ab then
		    let (spill_i, layout') = spill_reg layout r2 in
		    let (il, cbs) = 
		      cont (YCconst YCint, add_id layout' x (Rloc r2))
		    in
		    (spill_i @
		     (ArithBin (ab, Reg r2, raw1) ::
		      il), cbs)
		  else
		    let (r,spill_i,layout') = get_reg layout in
		    let ((raw2,clist2) as e2',_) = 
		      cogen_val (set_layout env layout') e2 in
		    let (il, cbs) = cont (YCconst YCint, add_id layout' x (Rloc r)) 
		    in
		    (spill_i @ 
		     (Mov (Reg r, e1') ::
	       	      ArithBin (ab, Reg r, raw2) :: 
	      	      il), cbs)
	      |	_,_ ->
		  let (r,spill_i,layout') = get_reg layout in
		  let (il, cbs) = cont (YCconst YCint, add_id layout' x (Rloc r))
		  in
		  (spill_i @
		   (Mov (Reg r, e1') ::
		    ArithBin (ab, Reg r, raw2) ::
		    il), cbs)
	      end
	| Bind_loc r ->
	    let (spill_i, layout') = spill_reg layout r in
            let ((raw1,clist1) as e1', _) = 
	      cogen_val (set_layout env layout') e1 in
	    let ((raw2,clist2) as e2',_) = 
	      cogen_val (set_layout env layout') e2 in
	    if clist1 != [] or clist2 != [] then die "arithbin: coerced args";
	    let (il, cbs) = cont  (YCconst YCint, layout') in
	    (spill_i @ 
	     Mov (Reg r, raw raw1) ::
	     ArithBin (ab, Reg r, raw2) ::
	     il, cbs)
	end
    |	_ -> die "cogen_simple_rhs: arithbin: not two arguments"
  in
  (* Compute a unary arithmetic operation on es and continue with cont *)
  let arithun au es =
    match es with
      [e1] ->
	begin
	  match dest with 
	  Bind_id x ->
	    if is_dead deadvars x e1 then
	      let ((raw1,clist1), _) = cogen_val env e1 in
	      singlei (ArithUn (au, raw1))
		(YCconst YCint, replace_id layout x e1) cont
	    else
	      let (r, spill_i, layout') = get_reg layout in
	      let ((raw1,clist1) as e1', _) = 
		cogen_val (set_layout env layout') e1 in
	      let (il, cbs) = cont (YCconst YCint, add_id layout' x (Rloc r)) 
	      in
	      (spill_i @ 
               (Mov (Reg r, e1') ::
	       	ArithUn (au, Reg r) :: 
	      	il), cbs)
	  | Bind_loc r ->
	      let (spill_i, layout') = spill_reg layout r in
	      let ((raw1,clist1) as e1',_) = 
		cogen_val (set_layout env layout') e1 in
	      let (il, cbs) = cont (YCconst YCint, layout') in
	      (spill_i @
	       (Mov (Reg r, e1') ::
		ArithUn (au, Reg r) ::
		il), cbs)
	end
    | _ -> die "cogen_prim: arithun: unary operator does not have 1 argument"
  in
  (* Compute a comparison operation (=,<,>, etc) on es and continue with
     cont.
  *)
  let compare_op cmp es =
    match es with
      [e1;e2] ->
	begin
	  let ((raw1, clist1) as e1', _) = 
	    cogen_val env e1 in
	  let ((raw2, clist2) as e2',_) = 
	    cogen_val env e2 in
	  if clist1 != [] or clist2 != [] then die "compare_op: coerced args";
	  let cont' layout rdest = 
	    let (il, cbs) = cont (ycbool, layout) in
	    (* Clear operand and then move condition code in *)
	    (Mov (Reg rdest, 
		 coerce (raw (Tag 0)) 
		   (Tosum (Csum {sum_tags = [0;1]; sum_vt= None}))) ::
	    Setcc (cmp, Reg rdest) :: il, cbs)
	  in
	  match dest with
	    Bind_id x ->
	      let (r, spill_i, layout') = 
		get_reg_from layout [Eax;Ebx;Ecx;Edx] in
	      let (il,cbs) = cont' (add_id layout' x (Rloc r)) r in
	      if valid_binops raw1 raw2 then
		(Cmp (raw1, raw2) :: (spill_i @ il), cbs)
	      else
		(Mov (Reg r, e1') :: Cmp (Reg r, raw2) :: (spill_i @ il), cbs)
	  | Bind_loc r ->
	      begin 
		(* The destination of setcc must be one of Eax, Ebx, Ecx, Edx.
		   If r is not one of these, it is a pain in the butt.
		*)
		match r with
		  Eax | Ebx | Ecx | Edx ->
		    let (spill_i, layout') = spill_reg layout r in
		    let (il,cbs) = cont' layout' r in
		    if valid_binops raw1 raw2 then
		      (Cmp (raw1, raw2) :: (spill_i @ il), cbs)
		    else
		      (Mov (Reg r, e1') :: 
		       Cmp (Reg r, raw2) :: (spill_i @ il), cbs)
		| r ->
		    let (rcomp, spill_i, layout') = 
		      get_reg_from layout [Ebx;Ecx;Edx;Eax] in
		    let (spill2, layout'') = spill_reg layout' r in
		    let (il, cbs) = cont (ycbool, layout'') in
		    let il = 	    
		      Mov (Reg rcomp, 
			   coerce (raw (Tag 0)) 
			     (Tosum (Csum {sum_tags = [0;1]; sum_vt= None})))::
		      Setcc (cmp, Reg rcomp) ::  
		      (spill2 @
		       (Mov (Reg r, raw (Reg rcomp)) :: il))
		    in
		    if valid_binops raw1 raw2 then
		      (Cmp (raw1, raw2) :: (spill_i @ il), cbs)
		    else
		      (Mov (Reg r, e1') :: 
		       Cmp (Reg r, raw2) :: (spill_i @ il), cbs)

	      end
	end
    | _ -> die "cogen_prim: comparison operator does not have 2 operands"
  in
  (* Compute the multiply or divide operation md on arguments es, moving 
     the result from register resreg (either Eax or Edx) to rdest.
  *)
  let arithmd md resreg es =
    match es with
      [e1; e2] ->
	let (spill1, layout') = spill_reg layout Eax in
	let (spill2, layout'') = spill_reg layout' Edx in
	let env = set_layout env layout'' in
	let ((raw1,clist1) as e1', _) = cogen_val env e1 in
	let ((raw2,clist2) as e2', _) = cogen_val env e2 in
	if clist1 != [] or clist2 != [] then die "compare_op: coerced args";
	let finish layout3 = 
	  let (il, cbs) = cont (YCconst YCint, layout3) in
	       (spill1 @ (spill2 @
			  (Mov (Reg Eax, e1') ::
			   ArithMD (md, raw2) ::
			   il)), cbs)
	in
	 begin
	   match dest with 
	     Bind_id x -> finish (add_id layout'' x (Rloc Eax))
	   | Bind_loc r ->
	       if r = resreg then
		 finish layout''
	       else
		 let (spill3, layout''') = spill_reg layout'' r in
		 let (il, cbs) = cont (YCconst YCint, layout''') in
		 (spill1 @ (spill2 @ (spill3 @
			    (Mov (Reg Eax, e1') ::
			     ArithMD (md, raw2) ::
			     Mov (Reg r, raw (Reg resreg)) ::
			     il))), cbs)
	 end
    | _ -> die "cogen_prim: arith multiply/divide operator does not have 2 operands"
  in
  (* Compute a shift/rotate operation on es and continue with cont *)
  let arithsr sr es =
    match es with
      [e1;e2] ->
	begin
	  let cont' layout raw1 raw2 xopt =
	    (match raw2 with
	      Immed i ->
		let layout' =
		  match xopt with
		    None -> layout
		  |	Some x -> replace_id layout x e1 
		in
		let (il, cbs) = cont (YCconst YCint, layout') in
	  	(ArithSR (sr, raw1, Some i) ::
	  	 il, cbs)
	    | _ ->
		let (spill1, layout') = spill_reg layout Ecx in
		let layout'' = 
		  match xopt with
		    None -> layout'
		  | Some x -> replace_id layout' x e1
		in
		let (il, cbs) = cont (YCconst YCint, layout'') in
	  	(spill1 @
		 (Mov (Reg Ecx, raw raw2) ::
	  	  ArithSR (sr, raw1, None) ::
	  	  il), cbs)
	     )    
	  in
	  match dest with 
	    Bind_id x ->
	      if is_dead deadvars x e1 then
		let ((raw1, clist1) as e1', _) = cogen_val env e1 in
		let ((raw2, clist2) as e2', _) = cogen_val env e2 in
		if clist1 != [] or clist2 != [] then 
		  die "shift_op: coerced args";
		cont' layout raw1 raw2 (Some x)
	      else
		let (r,spill1, layout') = get_reg layout in
		let env = set_layout env layout' in
		let ((raw1, clist1) as e1', _) = cogen_val env e1 in
		let ((raw2, clist2) as e2', _) = cogen_val env e2 in
		if clist1 != [] or clist2 != [] then 
		  die "shift_op: coerced args";
		let (il, cbs) = cont' layout' raw1 raw2 (Some x) in
		(spill1 @ (Mov (Reg r, e1') :: il), cbs)
	  | Bind_loc r ->
	      let (spill_i, layout') = spill_reg layout r in
	      let ((raw1, clist1) as e1', _) = cogen_val env e1 in
	      let ((raw2, clist2) as e2', _) = cogen_val env e2 in
	      if clist1 != [] or clist2 != [] then 
		die "shift_op: coerced args";
	      let (il, cbs) = cont' layout' raw1 raw2 None in
	      (spill_i @ (Mov (Reg r, e1') :: il), cbs)
	end
    | _ -> die "cogen_prim: arithun: unary operator does not have 1 argument"
  in
  let cogen_constant genop (c,layout) cont =
    match dest with
      Bind_id x ->
	begin
	  match (get_free_reg layout) with
	    None ->
    	      let (loc,layout) = add_id_on_stack layout x in
	      singlei (Mov (loc2genop frame loc, raw genop)) (c, layout) cont
	  | Some r ->
	      singlei (Mov (Reg r, raw genop)) 
		(c, add_id layout x (Rloc r)) cont
	end
    | Bind_loc r ->
	let (spill_i, layout') = spill_reg layout r in
	let (il, cbs) = cont (c, layout') in
	(spill_i @ (Mov (Reg r, raw genop) :: il), cbs)
  in
  match p with
    Yunit -> cogen_constant (Tag 0) (YCunit, layout) cont
  (* ints *)
  | Yint_i i ->
      cogen_constant (Immed i) (YCconst YCint, layout) cont
  | Yneg_i ->
      arithun Neg es
  | Yplus_i ->
      arithbin Add es
  | Yminus_i ->
      arithbin Sub es
  | Ytimes_i ->
      arithmd Imul1 Eax es
  | Ydiv_i ->
      arithmd Idiv Eax es
  | Ymod_i ->
      arithmd Idiv Edx es
  | Yeq_i ->
      compare_op Eq es
  | Ylt_i ->
      compare_op Less es
  | Ygt_i ->
      compare_op Greater es
  | Yneq_i ->
      compare_op NotEq es
  | Ylte_i ->
      compare_op LessEq es
  | Ygte_i ->
      compare_op GreaterEq es
  | Ynot_i ->
      arithun Not es
  | Yleft_shift_i ->
      arithsr Sal es
  | Yarith_right_shift_i ->
      arithsr Sar es
  | Ylogic_right_shift_i ->
      arithsr Shr es
  | Yand_i ->
      arithbin And es 
  | Yor_i ->
      arithbin Or es
  | Yxor_i ->
      arithbin Xor es
  | Yint_to_string -> raise (Cogen_unimplemented "Yint_to_string")
  | Yint_to_real -> raise (Cogen_unimplemented "Yint_to_real")
  | _ -> raise (Cogen_unimplemented 
		    "cogen_prim_rhs primop not implemented")

