(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker,                       *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

(* Code Generation of Values *)
open Utilities
open Ident
open Variable
open Tal
open Talcon
open Yprim
open Ycon
open Ycalculus
open Format
open Talpp
open Cogen_error
open Cogen_con
open Cogen_util

(* cogen_val: takes a code generation env and an anormal form value v
   Results:
   (genop coerce, Ycon.ycon) 
   where the ycon is normalized
*)
let rec cogen_val
    (((delta, gamma, kappa) as ctxt,
      ((ra, regs, frame, ret, deadvars) as layout),
      ((rho1, rho2) as taldelta)) as env : env) 
	v : (genop coerce * ycon) =
  match v with
    Yvar x ->
      (raw (loc2genop frame (get_loc layout x)),
       handle (fun _ -> Dict.lookup gamma x)
	  ("cogen_val: var "^(string_of_int x)^" not found in gamma"))
  (* Primitives which are values *)
  | Yprim(Yunit, _, _) -> (raw (Tag 0), YCunit) 
  | Yprim(Yany, cs, es) -> 
      (match (es, cs) with
	([v], [c]) -> 
	  let (v',c') = cogen_val env v in
	  let cnorm = ynorm kappa c in
	  (coerce v' (Toany (cogen_con cnorm)), cnorm) 
      |	_ -> raise (Cogen "Yany value translation: improper arguments"))
  | Yprim(Yint_i i, [], []) -> (raw (Immed i), YCconst YCint)
  | Yprim(Yreal_r s, cs, es) -> 
      raise (Cogen_unimplemented "float value translation")
  | Yprim(Yempty_a, _ , _) ->
      raise (Cogen_unimplemented "empty array translation")
  | Yprim(Yempty_v, _, _) -> 
      raise (Cogen_unimplemented "empty vector translation")
  | Yprim(Ystring_s s, _, _) -> 
      raise (Cogen_unimplemented "string value translation")

  | Ytyapp (f, cs) ->
      let normcs = List.map (ynorm kappa) cs in
      let (f', normcf) = (cogen_val env f) in
      let normresc = 
	match normcf with
	  YCabs (vks,argcs,resc) ->
	    (* Build the substitution kappa' binding abstracted type variables
	       to the normalized constructors which are applied.
	       Return also the remaining type variable bindings. *)
	    let rec new_vks vks cs kappa =
	      match (vks, cs) with
		((v,_)::vks', c::cs') -> 
		  new_vks vks' cs' (Dict.insert kappa v c)
	      |	(vks',[]) ->
		  (vks', kappa)
	      |	_ -> die "cogen val: Ytyapp: too many constructor arguments"
	    in
	    let (vks', kappa') = new_vks vks normcs kappa in
	    YCabs (vks', List.map (ynorm kappa') argcs, ynorm kappa' resc)
	| _ -> die "cogen val: Ytyapp: application of non type abstraction"
      in
      (List.fold_left
	 (fun res c -> coerce res (Tapp (cogen_con c))) f' normcs, normresc)
  | Ypack (e, v, k, c1, c2) ->
      let normc1 = ynorm kappa c1 in
      let normc2 = ynorm kappa c2 in
      let (e', _) = cogen_val env e in
      (coerce e' (Pack (Cexist(v, cogen_kind k, cogen_con normc1), 
			    cogen_con normc2)),
       YCexist (v,k,normc1))
  | Yfold (c, c1, c2, e) ->
      raise (Cogen_unimplemented "fold value translation")
  | Yunfold (k, c1, c2, e) ->
      raise (Cogen_unimplemented "unfold value translation")
  | Yroll (c, e) ->
      let normc = ynorm kappa c in
      let (e',c') = cogen_val env e in
      (coerce e' (Roll (cogen_con normc)), normc)
  | Yunroll e ->
      let (e',c') = cogen_val env e in
      (match c' with
 	YCmu (v,c) ->
	  (coerce e' Unroll, ynorm (Dict.insert kappa v c') c)
      |	_ -> die "cogen val: Yunroll: not mu type")
  | Ytag (i, css, []) ->
      (* Assuming this is a tag value, not a tuple *)
      let normc = YCsum (List.map (List.map (ynorm kappa)) css) in
      (coerce (raw (Tag i)) (Tosum (cogen_con normc)), normc) 
  | _ -> (dumpye v; die "cogen_val: non-value")

and cogen_vals env vs =
  List.map (cogen_val env) vs
