/* Standard Library for TAL Code
 *
 * Copyright Greg Morrisett, Neal Glew, Dan Grossman
 *   January 1999, all rights reserved
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/stat.h>

extern void *GC_malloc(int size);
extern void *GC_malloc_atomic(int size);

typedef struct str_internal {int size; char *chars;} *string;
typedef char *Cstring;
typedef struct arr_internal {int size; void *elts;} *array;
typedef void *(*gen_fun)(void *,int);

/***************************** I/O *******************************/


extern FILE *tal_stderr;
extern FILE *tal_stdin;
extern FILE *tal_stdout;

static void check_fd(FILE *fd) {
  if(!fd) {
    fprintf(stderr,"Attempt to access null file descriptor.\n");
    exit(255);
  }
}

void fprint_int    (FILE *fd, int x)    { 
  check_fd(fd);
  if(fprintf(fd, "%d", x) < 0)
    fprintf(stderr,"File write failed (non-fatal)\n");
}
void fprint_uint    (FILE *fd, int x)    { 
  check_fd(fd);
  if(fprintf(fd, "%u", x) < 0)
    fprintf(stderr,"File write failed (non-fatal)\n");
}
void fprint_char   (FILE *fd, char c)   { 
  check_fd(fd);
  if (putc(c, fd) == EOF)
    fprintf(stderr,"File write failed (non-fatal)\n");
}
void fprint_string (FILE *fd, string s) { 
  check_fd(fd);
  if (fwrite(s->chars, 1, s->size, fd) < s->size) 
    fprintf(stderr,"File write failed (non-fatal)\n");
}
void fprint_float  (FILE *fd, float f) {
  check_fd(fd);
  if (fprintf(fd, "%G", (double)f) < 0)
	fprintf(stderr,"File write failed (non-fatal)\n");
}
void fprint_double (FILE *fd, double d) {
  check_fd(fd);
  if (fprintf(fd, "%G", d) < 0)
	fprintf(stderr,"File write failed (non-fatal)\n");
}
void fprint_newline(FILE *fd)           { 
  check_fd(fd);
  if (putc('\n', fd) == EOF)
    fprintf(stderr,"File write failed (non-fatal)\n");
}
int f_close(FILE *fd) {
  if(!fd) {
    fprintf(stderr,"Attempt to close null file descriptor.\n");
    exit(255);
  }
  return fclose(fd);
}
int f_string_read(FILE *fd, string dest, int dest_offset, int max_count) {
  check_fd(fd);
  if(dest_offset + max_count > dest->size) {
    fprintf(stderr,"Attempt to read off end of string.\n");
    exit(255);
  }
  return fread((dest->chars)+dest_offset, 1, max_count, fd);
}
int f_length(const char *filename) {
  struct stat statbuf;
  if (stat(filename,&statbuf)) {
    fprintf(stderr, "Cannot stat %s", filename);
    return 0;
  }
  else {
    return statbuf.st_size;
  }
}

/***************************** Timing *******************************/

/* This is ad hoc and doesn't work under Unix, but for now...
 */

#ifdef __linux__
#  define _ftime ftime
#  define _timeb timeb
#  include <sys/timeb.h>
#else

typedef long time_t;
struct _timeb {
        time_t time;
        unsigned short millitm;
        short timezone;
        short dstflag;
        };

#endif

int int_time()
{
  struct _timeb tstr;
  long result;

  /* places secs in "time field", msecs in "millitm" field */
  _ftime( &tstr );
  
  /* get secs */
  result = tstr.time;

  /* free up top ten 10 bits (to avoid overflow of following multiplication) */
  result = result & 0x003fffff;

  /* convert from secs to msecs (now uses all 32 bits) */
  result = result * 1000;
  
  /* add in msecs */
  result = result + tstr.millitm;

  return result;
}

/********************** Runtime subversion *********************/
/* tal_start is a pain so we don't use it.  Just duplicate code here. */


Cstring string_to_Cstring(string s) {
  char *str=GC_malloc(s->size+1);
  char *chars = s->chars;

  int i;
  for(i=0; i<s->size; i++) str[i]=chars[i];
  str[s->size]='\0';
  return str;
}

string  Cstring_to_string(Cstring s) {
  string str;
  char *data;
  int size=(s?strlen(s):0);
  str=GC_malloc(sizeof(struct str_internal));
  data = GC_malloc_atomic(size);
  str->size=size;
  str->chars = data;
  while(--size>=0) data[size]=s[size];

  return str;
  /* We must make a copy in case C code frees the string or mangles it. 
     Otherwise we could get away with str->chars = s;
  */
}

array new_array(int,void *);
extern int tal_argc;
extern char **tal_argv;

array std_args() {
  array result = new_array(tal_argc,(void *)0);
  void **data = result->elts;
  int arg = tal_argc;
  while(arg--) 
    data[arg]=Cstring_to_string(tal_argv[arg]);
  return result;
}

/************************Arrays***********************/
array   new_array(int size, void *init) { 
  void **a=GC_malloc(size*4);
  array na = GC_malloc(sizeof(struct arr_internal));
  na->size=size;
  na->elts=a;
  if (init) {
    void **a_end = a+size;
    for(;a<a_end;a++)
      *a=init;
  }
  return na;
}

array   new_array_f(int size, gen_fun f, void *env) {
  int i;
  void ** a = GC_malloc(size*4);
  array na = GC_malloc(sizeof(struct arr_internal));
  na->size=size;
  na->elts = a;

  for(i=0; i<size; i++)
    a[size]=(*f)(env,i);

  return na;
}

array   new_array1(int size) {
  array na=GC_malloc(sizeof(struct arr_internal));
  void *a=GC_malloc_atomic(size);
  na->size = size;
  na->elts = a;
  return na;
}

array   new_array2(int size) {
  array na=GC_malloc(sizeof(struct arr_internal));
  void *a=GC_malloc_atomic(2*size);
  na->size = size;
  na->elts = a;
  return na;
} 

array   new_array4(int size) {
  array na=GC_malloc(sizeof(struct arr_internal));
  void *a=GC_malloc_atomic(4*size);
  na->size = size;
  na->elts = a;
  return na;
} 

array   new_array8(int size) {
  array na=GC_malloc(sizeof(struct arr_internal));
  void *a=GC_malloc_atomic(8*size);
  na->size = size;
  na->elts = a;
  return na;
} 

/**** pointer comparison and hashing ****/

int ptr_compare(void *ptr1, void *ptr2) {
  if (ptr1 == ptr2)
    return 0;
  else if ((unsigned int)ptr1 < (unsigned int)ptr2) 
    return -1;
  else
    return 1;
}

int hash(void *ptr) {
  return (int)ptr;
}

/**** Loader initialization ****/

#ifdef __linux__
#define SYMNAME(s) s
#else
#define SYMNAME(s) "_" ## s
#endif

void
stdlib_init_loader_syms(void (*register_sym)(long,long))
{
  /* nothing for now */
}

/***************************** Temporary *******************************/

/* This stuff should really go into a standard library or utility file
 * somewhere else.  But its here for now.
 */

void print_time(int start_time, int end_time, int iterations)
{
  printf("Time: %d msec\n", (end_time - start_time) / iterations);
}
/*
void get_input(string out)
{
  char c;

  fprint_string(tal_stdout, out);
  printf("(Hit <RETURN> to continue, 'q' + <RETURN> to quit)");
  fflush(stdout);
  c = getchar();
  if (c == 'q')
    exit(0);
}
*/
/* EOF: stdlib.c */

/*** CODE BELOW GENERATED AUTOMATICALLY BY genCinit.exe ****/

typedef struct rep_internal {int size; char chars[1];} *rep;


/*** getenv ***/
typedef struct { int len; char s[201]; } rep201;
static rep201 ctx_getenv = { 201, { 15,115,49,0,115,50,0,101,49,0,101,50,0,67,115,116,114,105,110,103,63,0,99,36,48,0,115,0,9,83,67,77,73,101,76,9,86,1,86,3,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,9,51,86,7,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,115,65,115,61,86,5,106,5,71,55,109,0,67,69,65,1,0,73,67,75,58,77,65,115,79,109,0,81,69,0,73,67,83,65,7,3,85,65,5,3,87,65,3,1,89,65,1,1,91,70,93,66,42,3,95,33,1,1,97,0,1,99 }};
static struct str_internal str_getenv = { 6, "getenv" };

/*** system ***/
static rep201 ctx_system = { 201, { 15,115,49,0,115,50,0,101,49,0,101,50,0,67,115,116,114,105,110,103,63,0,99,36,48,0,115,0,9,83,67,77,73,101,76,9,86,1,86,3,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,9,51,86,7,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,115,65,115,61,86,5,106,5,71,55,109,0,67,69,65,11,0,73,67,75,58,77,65,115,79,109,0,81,69,0,73,67,83,65,7,3,85,65,5,3,87,65,3,1,89,65,1,1,91,70,93,66,42,3,95,33,1,1,97,0,1,99 }};
static struct str_internal str_system = { 6, "system" };

/*** tal_stdin ***/
typedef struct { int len; char s[35]; } rep35;
static rep35 ctx_tal_stdin = { 35, { 3,70,73,76,69,63,0,1,15,76,1,70,1,66,42,3,3,33,1,1,5,0,70,7,66,42,3,9,33,1,1,11,0,1,13 }};
static struct str_internal str_tal_stdin = { 9, "tal_stdin" };

/*** tal_stdout ***/
static rep35 ctx_tal_stdout = { 35, { 3,70,73,76,69,63,0,1,15,76,1,70,1,66,42,3,3,33,1,1,5,0,70,7,66,42,3,9,33,1,1,11,0,1,13 }};
static struct str_internal str_tal_stdout = { 10, "tal_stdout" };

/*** tal_stderr ***/
static rep35 ctx_tal_stderr = { 35, { 3,70,73,76,69,63,0,1,15,76,1,70,1,66,42,3,3,33,1,1,5,0,70,7,66,42,3,9,33,1,1,11,0,1,13 }};
static struct str_internal str_tal_stderr = { 10, "tal_stderr" };

/*** fopen ***/
typedef struct { int len; char s[212]; } rep212;
static rep212 ctx_fopen = { 212, { 17,115,49,0,115,50,0,101,49,0,101,50,0,67,115,116,114,105,110,103,63,0,99,36,48,0,115,0,70,73,76,69,63,0,9,83,67,77,73,105,76,9,86,1,86,3,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,9,51,86,7,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,58,1,65,115,67,115,61,76,15,86,5,106,5,75,55,109,0,69,71,65,73,0,77,67,79,58,81,67,115,83,109,0,85,71,0,77,67,87,65,7,3,89,65,5,3,91,65,3,1,93,65,1,1,95,70,97,66,42,3,99,33,1,1,101,0,1,103 }};
static struct str_internal str_fopen = { 5, "fopen" };

/*** f_length ***/
static rep201 ctx_f_length = { 201, { 15,115,49,0,115,50,0,101,49,0,101,50,0,67,115,116,114,105,110,103,63,0,99,36,48,0,115,0,9,83,67,77,73,101,76,9,86,1,86,3,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,9,51,86,7,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,115,65,115,61,86,5,106,5,71,55,109,0,67,69,65,11,0,73,67,75,58,77,65,115,79,109,0,81,69,0,73,67,83,65,7,3,85,65,5,3,87,65,3,1,89,65,1,1,91,70,93,66,42,3,95,33,1,1,97,0,1,99 }};
static struct str_internal str_f_length = { 8, "f_length" };

/*** f_close ***/
typedef struct { int len; char s[198]; } rep198;
static rep198 ctx_f_close = { 198, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,101,76,9,86,1,86,3,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,9,51,86,7,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,115,65,115,61,86,5,106,5,71,55,109,0,67,69,65,11,0,73,67,75,58,77,65,115,79,109,0,81,69,0,73,67,83,65,7,3,85,65,5,3,87,65,3,1,89,65,1,1,91,70,93,66,42,3,95,33,1,1,97,0,1,99 }};
static struct str_internal str_f_close = { 7, "f_close" };

/*** fprint_int ***/
typedef struct { int len; char s[199]; } rep199;
static rep199 ctx_fprint_int = { 199, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,103,76,9,52,86,1,86,3,115,7,84,70,3,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,11,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,11,51,86,7,109,1,9,65,53,0,55,67,57,58,59,7,64,5,61,58,3,63,58,1,65,115,67,115,61,86,5,106,5,73,55,109,0,69,71,0,75,67,77,58,79,67,115,81,109,0,83,71,0,75,67,85,65,7,3,87,65,5,3,89,65,3,1,91,65,1,1,93,70,95,66,42,3,97,33,1,1,99,0,1,101 }};
static struct str_internal str_fprint_int = { 10, "fprint_int" };

/*** fprint_char ***/
static rep199 ctx_fprint_char = { 199, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,103,76,9,52,86,1,86,3,115,7,84,70,3,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,11,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,11,51,86,7,109,1,9,65,53,0,55,67,57,58,59,7,64,5,61,58,3,63,58,1,65,115,67,115,61,86,5,106,5,73,55,109,0,69,71,0,75,67,77,58,79,67,115,81,109,0,83,71,0,75,67,85,65,7,3,87,65,5,3,89,65,3,1,91,65,1,1,93,70,95,66,42,3,97,33,1,1,99,0,1,101 }};
static struct str_internal str_fprint_char = { 11, "fprint_char" };

/*** fprint_float ***/
typedef struct { int len; char s[200]; } rep200;
static rep200 ctx_fprint_float = { 200, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,105,76,9,51,86,1,86,3,115,7,84,52,70,13,66,42,3,15,86,11,33,1,1,17,1,19,66,70,21,82,86,13,83,25,70,27,82,49,70,31,66,118,25,33,42,3,35,33,1,1,37,0,70,39,82,42,5,29,41,33,1,1,43,0,69,13,7,11,45,70,47,82,42,7,23,49,19,33,1,1,51,0,69,11,5,11,53,86,7,109,1,9,65,55,0,57,67,59,58,61,7,64,5,63,58,3,65,58,1,67,115,69,115,63,86,5,106,5,75,57,109,0,71,73,0,77,67,79,58,81,69,115,83,109,0,85,73,0,77,67,87,65,7,3,89,65,5,3,91,65,3,1,93,65,1,1,95,70,97,66,42,3,99,33,1,1,101,0,1,103 }};
static struct str_internal str_fprint_float = { 12, "fprint_float" };

/*** fprint_double ***/
static rep200 ctx_fprint_double = { 200, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,105,76,9,54,86,1,86,3,115,7,84,52,70,13,66,42,3,15,86,11,33,1,1,17,1,19,66,70,21,82,86,13,83,25,70,27,82,49,70,31,66,118,25,33,42,3,35,33,1,1,37,0,70,39,82,42,5,29,41,33,1,1,43,0,69,13,7,11,45,70,47,82,42,7,23,49,19,33,1,1,51,0,69,11,5,11,53,86,7,109,1,9,65,55,0,57,67,59,58,61,7,64,5,63,58,3,65,58,1,67,115,69,115,63,86,5,106,5,75,57,109,0,71,73,0,77,67,79,58,81,69,115,83,109,0,85,73,0,77,67,87,65,7,3,89,65,5,3,91,65,3,1,93,65,1,1,95,70,97,66,42,3,99,33,1,1,101,0,1,103 }};
static struct str_internal str_fprint_double = { 13, "fprint_double" };

/*** fprint_string ***/
static rep199 ctx_fprint_string = { 199, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,115,0,99,36,48,0,9,83,67,73,77,103,76,9,84,86,11,83,5,70,7,82,49,70,11,66,118,5,13,42,3,15,33,1,1,17,0,70,19,82,42,5,9,21,33,1,1,23,0,69,11,5,3,25,86,1,86,3,115,31,52,70,35,66,42,3,37,86,13,33,1,1,39,1,41,66,70,43,82,70,27,82,42,7,45,47,41,33,1,1,49,0,69,13,7,3,51,86,7,109,1,33,65,53,0,55,67,57,58,59,31,64,29,61,58,27,63,58,1,65,115,67,115,61,86,5,106,5,73,55,109,0,69,71,0,75,67,77,58,79,67,115,81,109,0,83,71,0,75,67,85,65,7,3,87,65,5,3,89,65,3,1,91,65,1,1,93,70,95,66,42,3,97,33,1,1,99,0,1,101 }};
static struct str_internal str_fprint_string = { 13, "fprint_string" };

/*** fprint_newline ***/
typedef struct { int len; char s[196]; } rep196;
static rep196 ctx_fprint_newline = { 196, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,101,76,9,86,1,86,3,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,9,51,86,7,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,115,65,115,61,86,5,106,5,71,55,109,0,67,69,0,73,67,75,58,77,65,115,79,109,0,81,69,0,73,67,83,65,7,3,85,65,5,3,87,65,3,1,89,65,1,1,91,70,93,66,42,3,95,33,1,1,97,0,1,99 }};
static struct str_internal str_fprint_newline = { 14, "fprint_newline" };

/*** fflush ***/
static rep196 ctx_fflush = { 196, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,101,76,9,86,1,86,3,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,9,51,86,7,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,115,65,115,61,86,5,106,5,71,55,109,0,67,69,0,73,67,75,58,77,65,115,79,109,0,81,69,0,73,67,83,65,7,3,85,65,5,3,87,65,3,1,89,65,1,1,91,70,93,66,42,3,95,33,1,1,97,0,1,99 }};
static struct str_internal str_fflush = { 6, "fflush" };

/*** fgetc ***/
static rep198 ctx_fgetc = { 198, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,101,76,9,86,1,86,3,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,9,51,86,7,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,115,65,115,61,86,5,106,5,71,55,109,0,67,69,65,11,0,73,67,75,58,77,65,115,79,109,0,81,69,0,73,67,83,65,7,3,85,65,5,3,87,65,3,1,89,65,1,1,91,70,93,66,42,3,95,33,1,1,97,0,1,99 }};
static struct str_internal str_fgetc = { 5, "fgetc" };

/*** ungetc ***/
static rep201 ctx_ungetc = { 201, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,103,76,9,52,86,1,86,3,115,7,84,70,3,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,11,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,11,51,86,7,109,1,9,65,53,0,55,67,57,58,59,7,64,5,61,58,3,63,58,1,65,115,67,115,61,86,5,106,5,73,55,109,0,69,71,65,3,0,75,67,77,58,79,67,115,81,109,0,83,71,0,75,67,85,65,7,3,87,65,5,3,89,65,3,1,91,65,1,1,93,70,95,66,42,3,97,33,1,1,99,0,1,101 }};
static struct str_internal str_ungetc = { 6, "ungetc" };

/*** f_string_read ***/
typedef struct { int len; char s[207]; } rep207;
static rep207 ctx_f_string_read = { 207, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,115,0,99,36,48,0,9,83,67,73,77,107,76,9,84,86,11,83,5,70,7,82,49,70,11,66,118,5,13,42,3,15,33,1,1,17,0,70,19,82,42,5,9,21,33,1,1,23,0,69,11,5,3,25,52,86,1,86,3,115,33,70,29,66,42,3,37,86,13,33,1,1,39,1,41,66,70,43,82,70,27,82,42,7,45,47,41,33,1,1,49,0,69,13,7,3,51,86,7,109,1,35,65,53,0,55,67,57,58,59,33,64,31,61,58,29,63,58,29,65,58,27,67,58,1,69,115,71,115,61,86,5,106,5,77,55,109,0,73,75,65,29,0,79,67,81,58,83,71,115,85,109,0,87,75,0,79,67,89,65,7,3,91,65,5,3,93,65,3,1,95,65,1,1,97,70,99,66,42,3,101,33,1,1,103,0,1,105 }};
static struct str_internal str_f_string_read = { 13, "f_string_read" };

/*** fprint_uint ***/
static rep199 ctx_fprint_uint = { 199, { 15,115,49,0,115,50,0,101,49,0,101,50,0,70,73,76,69,63,0,99,36,48,0,115,0,9,83,67,77,73,103,76,9,52,86,1,86,3,115,7,84,70,3,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,11,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,11,51,86,7,109,1,9,65,53,0,55,67,57,58,59,7,64,5,61,58,3,63,58,1,65,115,67,115,61,86,5,106,5,73,55,109,0,69,71,0,75,67,77,58,79,67,115,81,109,0,83,71,0,75,67,85,65,7,3,87,65,5,3,89,65,3,1,91,65,1,1,93,70,95,66,42,3,97,33,1,1,99,0,1,101 }};
static struct str_internal str_fprint_uint = { 11, "fprint_uint" };

/*** std_args ***/
typedef struct { int len; char s[218]; } rep218;
static rep218 ctx_std_args = { 218, { 13,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,9,83,67,77,73,113,86,1,86,3,115,3,84,52,70,9,66,42,3,11,86,9,33,1,1,13,1,15,66,70,17,82,86,11,83,21,70,23,82,49,70,27,66,118,21,29,42,3,31,33,1,1,33,0,70,35,82,42,5,25,37,33,1,1,39,0,69,11,7,7,41,70,43,82,42,7,19,45,15,33,1,1,47,0,69,9,5,7,49,86,7,109,1,5,65,51,0,53,67,55,58,57,3,64,1,59,115,61,115,59,70,43,66,118,21,67,42,3,69,33,1,1,71,0,70,73,82,42,5,25,75,33,1,1,77,0,69,11,7,7,79,86,5,106,5,83,53,109,0,63,65,65,81,0,85,67,87,58,89,61,115,91,109,0,93,65,0,85,67,95,65,7,3,97,65,5,3,99,65,3,1,101,65,1,1,103,70,105,66,42,3,107,33,1,1,109,0,1,111 }};
static struct str_internal str_std_args = { 8, "std_args" };

/*** string_to_Cstring ***/
static rep201 ctx_string_to_Cstring = { 201, { 15,115,49,0,115,50,0,101,49,0,101,50,0,115,0,99,36,48,0,67,115,116,114,105,110,103,63,0,9,83,67,73,77,101,84,86,9,83,3,70,5,82,49,70,9,66,118,3,11,42,3,13,33,1,1,15,0,70,17,82,42,5,7,19,33,1,1,21,0,69,9,5,1,23,86,1,86,3,115,29,52,70,33,66,42,3,35,86,11,33,1,1,37,1,39,66,70,41,82,70,25,82,42,7,43,45,39,33,1,1,47,0,69,11,7,1,49,86,7,109,1,31,65,51,0,53,67,55,58,57,29,64,27,59,58,25,61,115,63,115,59,76,13,86,5,106,5,71,53,109,0,65,67,65,69,0,73,67,75,58,77,63,115,79,109,0,81,67,0,73,67,83,65,7,3,85,65,5,3,87,65,3,1,89,65,1,1,91,70,93,66,42,3,95,33,1,1,97,0,1,99 }};
static struct str_internal str_string_to_Cstring = { 17, "string_to_Cstring" };

/*** Cstring_to_string ***/
static rep201 ctx_Cstring_to_string = { 201, { 15,115,49,0,115,50,0,101,49,0,101,50,0,67,115,116,114,105,110,103,63,0,99,36,48,0,115,0,9,83,67,77,73,101,76,9,86,1,86,3,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,7,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,5,9,51,86,7,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,115,65,115,61,86,5,106,5,71,55,109,0,67,69,65,45,0,73,67,75,58,77,65,115,79,109,0,81,69,0,73,67,83,65,7,3,85,65,5,3,87,65,3,1,89,65,1,1,91,70,93,66,42,3,95,33,1,1,97,0,1,99 }};
static struct str_internal str_Cstring_to_string = { 17, "Cstring_to_string" };

/*** new_array ***/
typedef struct { int len; char s[233]; } rep233;
static rep233 ctx_new_array = { 233, { 15,97,0,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,11,52,83,67,77,73,121,52,86,1,86,3,86,5,115,7,84,70,1,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,9,11,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,7,11,51,86,9,109,1,9,65,53,0,55,67,57,58,59,7,64,5,61,58,3,63,58,1,65,115,67,115,61,70,3,66,118,23,73,42,3,75,33,1,1,77,0,70,79,82,42,5,27,81,33,1,1,83,0,69,13,9,11,85,86,7,106,5,89,55,109,0,69,71,65,87,0,91,67,93,58,95,67,115,97,109,0,99,71,0,91,67,101,65,9,5,103,65,7,5,105,65,5,3,107,65,3,3,109,65,1,1,111,70,113,66,42,3,115,33,1,1,117,0,1,119 }};
static struct str_internal str_new_array = { 9, "new_array" };

/*** new_array_f ***/
typedef struct { int len; char s[277]; } rep277;
static rep277 ctx_new_array_f = { 277, { 15,97,0,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,11,52,83,67,77,73,145,52,86,3,86,5,115,5,84,70,1,66,42,3,11,86,11,33,1,1,13,1,15,66,70,17,82,86,13,83,21,70,23,82,49,70,27,66,118,21,29,42,3,31,33,1,1,33,0,70,35,82,42,5,25,37,33,1,1,39,0,69,13,9,9,41,70,43,82,42,7,19,45,15,33,1,1,47,0,69,11,7,9,49,86,9,109,1,7,65,51,0,53,67,55,58,57,5,64,3,59,58,1,61,115,63,115,59,86,1,86,7,106,5,71,53,109,0,65,67,65,69,0,73,67,75,58,77,63,115,79,109,0,81,67,0,73,67,83,65,9,5,85,65,7,5,87,65,5,3,89,65,3,3,91,58,93,61,58,1,95,115,97,70,69,66,118,21,101,42,3,103,33,1,1,105,0,70,107,82,42,5,25,109,33,1,1,111,0,69,13,9,9,113,109,0,99,67,65,115,0,73,67,117,58,119,97,115,121,109,0,123,67,0,73,67,125,65,9,5,127,65,7,5,129,65,5,3,131,65,3,3,133,65,1,1,135,70,137,66,42,3,139,33,1,1,141,0,1,143 }};
static struct str_internal str_new_array_f = { 11, "new_array_f" };

/*** new_array1 ***/
typedef struct { int len; char s[190]; } rep190;
static rep190 ctx_new_array1 = { 190, { 13,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,9,83,67,77,73,99,52,86,1,86,3,115,5,84,70,1,66,42,3,11,86,9,33,1,1,13,1,15,66,70,17,82,86,11,83,21,70,23,82,49,70,27,66,118,21,29,42,3,31,33,1,1,33,0,70,35,82,42,5,25,37,33,1,1,39,0,69,11,7,9,41,70,43,82,42,7,19,45,15,33,1,1,47,0,69,9,5,9,49,86,7,109,1,7,65,51,0,53,67,55,58,57,5,64,3,59,58,1,61,115,63,115,59,86,5,106,5,69,53,109,0,65,67,65,43,0,71,67,73,58,75,63,115,77,109,0,79,67,0,71,67,81,65,7,3,83,65,5,3,85,65,3,1,87,65,1,1,89,70,91,66,42,3,93,33,1,1,95,0,1,97 }};
static struct str_internal str_new_array1 = { 10, "new_array1" };

/*** new_array2 ***/
typedef struct { int len; char s[222]; } rep222;
static rep222 ctx_new_array2 = { 222, { 13,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,9,83,67,77,73,117,52,86,1,86,3,115,5,84,70,1,66,42,3,11,86,9,33,1,1,13,1,15,66,70,17,82,86,11,83,21,70,23,82,49,70,27,66,118,21,29,42,3,31,33,1,1,33,0,70,35,82,42,5,25,37,33,1,1,39,0,69,11,7,9,41,70,43,82,42,7,19,45,15,33,1,1,47,0,69,9,5,9,49,86,7,109,1,7,65,51,0,53,67,55,58,57,5,64,3,59,58,1,61,115,63,115,59,50,70,69,66,118,21,71,42,3,73,33,1,1,75,0,70,77,82,42,5,25,79,33,1,1,81,0,69,11,7,9,83,86,5,106,5,87,53,109,0,65,67,65,85,0,89,67,91,58,93,63,115,95,109,0,97,67,0,89,67,99,65,7,3,101,65,5,3,103,65,3,1,105,65,1,1,107,70,109,66,42,3,111,33,1,1,113,0,1,115 }};
static struct str_internal str_new_array2 = { 10, "new_array2" };

/*** new_array4 ***/
static rep218 ctx_new_array4 = { 218, { 13,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,9,83,67,77,73,113,52,86,1,86,3,115,5,84,70,1,66,42,3,11,86,9,33,1,1,13,1,15,66,70,17,82,86,11,83,21,70,23,82,49,70,27,66,118,21,29,42,3,31,33,1,1,33,0,70,35,82,42,5,25,37,33,1,1,39,0,69,11,7,9,41,70,43,82,42,7,19,45,15,33,1,1,47,0,69,9,5,9,49,86,7,109,1,7,65,51,0,53,67,55,58,57,5,64,3,59,58,1,61,115,63,115,59,118,21,11,42,3,69,33,1,1,71,0,70,73,82,42,5,25,75,33,1,1,77,0,69,11,7,9,79,86,5,106,5,83,53,109,0,65,67,65,81,0,85,67,87,58,89,63,115,91,109,0,93,67,0,85,67,95,65,7,3,97,65,5,3,99,65,3,1,101,65,1,1,103,70,105,66,42,3,107,33,1,1,109,0,1,111 }};
static struct str_internal str_new_array4 = { 10, "new_array4" };

/*** new_array8 ***/
static rep222 ctx_new_array8 = { 222, { 13,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,9,83,67,77,73,117,52,86,1,86,3,115,5,84,70,1,66,42,3,11,86,9,33,1,1,13,1,15,66,70,17,82,86,11,83,21,70,23,82,49,70,27,66,118,21,29,42,3,31,33,1,1,33,0,70,35,82,42,5,25,37,33,1,1,39,0,69,11,7,9,41,70,43,82,42,7,19,45,15,33,1,1,47,0,69,9,5,9,49,86,7,109,1,7,65,51,0,53,67,55,58,57,5,64,3,59,58,1,61,115,63,115,59,56,70,69,66,118,21,71,42,3,73,33,1,1,75,0,70,77,82,42,5,25,79,33,1,1,81,0,69,11,7,9,83,86,5,106,5,87,53,109,0,65,67,65,85,0,89,67,91,58,93,63,115,95,109,0,97,67,0,89,67,99,65,7,3,101,65,5,3,103,65,3,1,105,65,1,1,107,70,109,66,42,3,111,33,1,1,113,0,1,115 }};
static struct str_internal str_new_array8 = { 10, "new_array8" };

/*** ptr_compare ***/
typedef struct { int len; char s[202]; } rep202;
static rep202 ctx_ptr_compare = { 202, { 15,97,0,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,11,52,83,67,77,73,105,86,1,86,3,86,5,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,9,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,7,9,51,86,9,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,58,1,65,115,67,115,61,86,7,106,5,73,55,109,0,69,71,65,11,0,75,67,77,58,79,67,115,81,109,0,83,71,0,75,67,85,65,9,5,87,65,7,5,89,65,5,3,91,65,3,3,93,65,1,1,95,70,97,66,42,3,99,33,1,1,101,0,1,103 }};
static struct str_internal str_ptr_compare = { 11, "ptr_compare" };

/*** hash ***/
static rep199 ctx_hash = { 199, { 15,97,0,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,11,52,83,67,77,73,103,86,1,86,3,86,5,115,5,84,52,70,11,66,42,3,13,86,11,33,1,1,15,1,17,66,70,19,82,86,13,83,23,70,25,82,49,70,29,66,118,23,31,42,3,33,33,1,1,35,0,70,37,82,42,5,27,39,33,1,1,41,0,69,13,9,9,43,70,45,82,42,7,21,47,17,33,1,1,49,0,69,11,7,9,51,86,9,109,1,7,65,53,0,55,67,57,58,59,5,64,3,61,58,1,63,115,65,115,61,86,7,106,5,71,55,109,0,67,69,65,11,0,73,67,75,58,77,65,115,79,109,0,81,69,0,73,67,83,65,9,5,85,65,7,5,87,65,5,3,89,65,3,3,91,65,1,1,93,70,95,66,42,3,97,33,1,1,99,0,1,101 }};
static struct str_internal str_hash = { 4, "hash" };

/*** int_time ***/
typedef struct { int len; char s[187]; } rep187;
static rep187 ctx_int_time = { 187, { 13,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,9,83,67,77,73,97,86,1,86,3,115,3,84,52,70,9,66,42,3,11,86,9,33,1,1,13,1,15,66,70,17,82,86,11,83,21,70,23,82,49,70,27,66,118,21,29,42,3,31,33,1,1,33,0,70,35,82,42,5,25,37,33,1,1,39,0,69,11,7,7,41,70,43,82,42,7,19,45,15,33,1,1,47,0,69,9,5,7,49,86,7,109,1,5,65,51,0,53,67,55,58,57,3,64,1,59,115,61,115,59,86,5,106,5,67,53,109,0,63,65,65,9,0,69,67,71,58,73,61,115,75,109,0,77,65,0,69,67,79,65,7,3,81,65,5,3,83,65,3,1,85,65,1,1,87,70,89,66,42,3,91,33,1,1,93,0,1,95 }};
static struct str_internal str_int_time = { 8, "int_time" };

/*** print_time ***/
typedef struct { int len; char s[194]; } rep194;
static rep194 ctx_print_time = { 194, { 13,115,49,0,115,50,0,101,49,0,101,50,0,99,36,48,0,115,0,9,83,67,77,73,103,52,86,1,86,3,115,5,84,70,1,66,42,3,11,86,9,33,1,1,13,1,15,66,70,17,82,86,11,83,21,70,23,82,49,70,27,66,118,21,29,42,3,31,33,1,1,33,0,70,35,82,42,5,25,37,33,1,1,39,0,69,11,7,9,41,70,43,82,42,7,19,45,15,33,1,1,47,0,69,9,5,9,49,86,7,109,1,7,65,51,0,53,67,55,58,57,5,64,3,59,58,1,61,58,1,63,58,1,65,115,67,115,59,86,5,106,5,73,53,109,0,69,71,0,75,67,77,58,79,67,115,81,109,0,83,71,0,75,67,85,65,7,3,87,65,5,3,89,65,3,1,91,65,1,1,93,70,95,66,42,3,97,33,1,1,99,0,1,101 }};
static struct str_internal str_print_time = { 10, "print_time" };
static void (*add_sym)(string sym,rep r,void *p);
void dyninit_stdlib(void *unused,void (*f)(string sym,rep r,void *p),
		    void *unused2) {
  add_sym = f;
  add_sym((string)&str_getenv,(rep)&ctx_getenv,(void *) getenv);
  add_sym((string)&str_system,(rep)&ctx_system,(void *) system);
  add_sym((string)&str_tal_stdin,(rep)&ctx_tal_stdin,(void *)&tal_stdin);
  add_sym((string)&str_tal_stdout,(rep)&ctx_tal_stdout,(void *)&tal_stdout);
  add_sym((string)&str_tal_stderr,(rep)&ctx_tal_stderr,(void *)&tal_stderr);
  add_sym((string)&str_fopen,(rep)&ctx_fopen,(void *) fopen);
  add_sym((string)&str_f_length,(rep)&ctx_f_length,(void *) f_length);
  add_sym((string)&str_f_close,(rep)&ctx_f_close,(void *) f_close);
  add_sym((string)&str_fprint_int,(rep)&ctx_fprint_int,(void *) fprint_int);
  add_sym((string)&str_fprint_char,(rep)&ctx_fprint_char,(void *) fprint_char);
  add_sym((string)&str_fprint_float,(rep)&ctx_fprint_float,(void *) fprint_float);
  add_sym((string)&str_fprint_double,(rep)&ctx_fprint_double,(void *) fprint_double);
  add_sym((string)&str_fprint_string,(rep)&ctx_fprint_string,(void *) fprint_string);
  add_sym((string)&str_fprint_newline,(rep)&ctx_fprint_newline,(void *) fprint_newline);
  add_sym((string)&str_fflush,(rep)&ctx_fflush,(void *) fflush);
  add_sym((string)&str_fgetc,(rep)&ctx_fgetc,(void *) fgetc);
  add_sym((string)&str_ungetc,(rep)&ctx_ungetc,(void *) ungetc);
  add_sym((string)&str_f_string_read,(rep)&ctx_f_string_read,(void *) f_string_read);
  add_sym((string)&str_fprint_uint,(rep)&ctx_fprint_uint,(void *) fprint_uint);
  add_sym((string)&str_std_args,(rep)&ctx_std_args,(void *) std_args);
  add_sym((string)&str_string_to_Cstring,(rep)&ctx_string_to_Cstring,(void *) string_to_Cstring);
  add_sym((string)&str_Cstring_to_string,(rep)&ctx_Cstring_to_string,(void *) Cstring_to_string);
  add_sym((string)&str_new_array,(rep)&ctx_new_array,(void *) new_array);
  add_sym((string)&str_new_array_f,(rep)&ctx_new_array_f,(void *) new_array_f);
  add_sym((string)&str_new_array1,(rep)&ctx_new_array1,(void *) new_array1);
  add_sym((string)&str_new_array2,(rep)&ctx_new_array2,(void *) new_array2);
  add_sym((string)&str_new_array4,(rep)&ctx_new_array4,(void *) new_array4);
  add_sym((string)&str_new_array8,(rep)&ctx_new_array8,(void *) new_array8);
  add_sym((string)&str_ptr_compare,(rep)&ctx_ptr_compare,(void *) ptr_compare);
  add_sym((string)&str_hash,(rep)&ctx_hash,(void *) hash);
  add_sym((string)&str_int_time,(rep)&ctx_int_time,(void *) int_time);
  add_sym((string)&str_print_time,(rep)&ctx_print_time,(void *) print_time);
}
