
(* cyclone.ml
   Contains macros used by Cyclone. *)

open Numtypes
open Tal
open Identifier


(* Macros to save and restore registers around function calls.

   Push4regs and pop4regs handle registers EAX, EBX, ECX, and EDX.

   (Push3regs reg) and (Pop3regs reg) save and restore EAX, EBX, ECX,
   and EDX, except for register reg.
*)

let push4regs =
   [  Push (Reg Eax, []);
      Push (Reg Ebx, []);
      Push (Reg Ecx, []);
      Push (Reg Edx, [])  ]

let pop4regs =
   [  Pop (Reg Edx);
      Pop (Reg Ecx);
      Pop (Reg Ebx);
      Pop (Reg Eax)  ]

let push3regs reg =
   match reg with
      Eax ->
	 [  Push (Reg Ebx, []);
	    Push (Reg Ecx, []);
	    Push (Reg Edx, [])  ]
    | Ebx ->
	 [  Push (Reg Eax, []);
	    Push (Reg Ecx, []);
	    Push (Reg Edx, [])  ]
    | Ecx ->
	 [  Push (Reg Eax, []);
	    Push (Reg Ebx, []);
	    Push (Reg Edx, [])  ]
    | Edx ->
	 [  Push (Reg Eax, []);
	    Push (Reg Ebx, []);
	    Push (Reg Ecx, [])  ]
    | _ -> invalid_arg "push3regs"

let pop3regs reg =
   match reg with
      Eax ->
	 [  Pop (Reg Edx);
	    Pop (Reg Ecx);
	    Pop (Reg Ebx)  ]
    | Ebx ->
	 [  Pop (Reg Edx);
	    Pop (Reg Ecx);
	    Pop (Reg Eax)  ]
    | Ecx ->
	 [  Pop (Reg Edx);
	    Pop (Reg Ebx);
	    Pop (Reg Eax)  ]
    | Edx ->
	 [  Pop (Reg Ecx);
	    Pop (Reg Ebx);
	    Pop (Reg Eax)  ]
    | _ -> invalid_arg "pop3regs"


(* Result register will be clobbered.  fun_call should not clobber any
   callee_save registers (ie EBX, ESI, EDI, EBP, and restore ESP).
*)
let preserve_regs result_reg fun_call =
  let save = 
    match result_reg with
      Some Eax -> [     Ecx; Edx]
    | Some Ecx -> [Eax;      Edx]
    | Some Edx -> [Eax; Ecx     ]
    | _        -> [Eax; Ecx; Edx] in (* All other regs are callee-save. *)
  let pop = List.map (fun x -> Pop (Reg x)) save in
  let push = List.rev (List.map (fun x -> Push (Reg x, [])) save) in
  let mov_result = 
    match result_reg with
      Some Eax -> []
    | None     -> []
    | Some reg -> [Mov (Reg reg, (Reg Eax, []))] in
  (push @ fun_call @ mov_result @ pop)

(* If we clobber a callee-save register then save that too. *)
let save_clobbered_reg reg instrs = 
  match reg with
    Eax | Ecx | Edx -> instrs (* Already saved. *)
  | _ -> ([Push (Reg reg,[])] @ instrs @ [Pop (Reg reg)])
    
(***** Label Arithmetic *****)

(* assert_template_labels
   Should raise an error if any label in the list is _not_ in the
   template section. For now, does nothing. *)

let assert_template_labels env label_list =
   () (**)(* Unimplemented *)

(* Labels of helper C functions defined in runtime/cyclonelib.c *)
let cg_start_label = (id_of_string "_CG_start")
let cg_dump_label = (id_of_string "_CG_dump")
let cg_end_label = (id_of_string "_CG_end")
let cg_mark_label = (id_of_string "_CG_mark")
let external_addr = (id_of_string "_external_addr")

let cyclone_labels = [ (cg_start_label, cempty);
			(cg_dump_label, cempty);
			(cg_end_label, cempty);
			(cg_mark_label, cempty);
			(external_addr, cempty) ]

(***** Macros *****)


(* CGSTART tv, con
   Allocate a new code generation region and return it in EAX.
   The arguments tv and con are used by the verifier only. *)

let cgstart_macro tv con =
  preserve_regs (Some Eax) [ Mov (Reg Eax, (Addr cg_start_label, []));
			     Call (Reg Eax, [])  ]


(* CGDUMP cgReg, tv, reg, lab

   Dump the template at label lab into the code generation region
   given by the ref cell cgReg.  The offset from the beginning of the
   code in the cg region to the beginning of the copied template is
   put into register reg.  The type variable tv is used by the
   verifier only; it is a BINDING OCCURRENCE. *)

let cgdump_macro cgReg tv reg lab =
  preserve_regs (Some reg) 
    [ Push (Addr lab, []);
      Push (Reg cgReg,[]);
      Mov (Reg Eax, (Addr cg_dump_label, []));
      Call (Reg Eax, []);
      ArithBin (Add, (Reg Esp), (Immed i32_8))  ]
    

(* CGFILL cgReg, reg1, tmplLab, holeLab, reg2

   cgReg is a register pointing to a region ref,
   reg1 is a register holding the offset of a copy of a template,
   tmplLab is the label of the original template,
   holeLab is the label of the hole of the original template,
   reg2 is a register containing the value to stuff into the hole
     of the copy *)

let cgfill_macro cgReg reg1 tmplLab holeLab reg2 =
   let disp = (holeLab -$ tmplLab -$ i32_3) in
   preserve_regs None 
     (save_clobbered_reg reg1 
	[ ArithBin (Add, (Reg reg1), (Immed disp));
	  Push (Reg reg1, []);
	  Push (Reg reg2, []);
	  Push (Reg cgReg,[]);
	  Mov (Reg Eax, (Addr cg_mark_label, []));
	  Call (Reg Eax, []);
	  ArithBin (Add, (Reg Esp), (Immed (int_to_int32 12)))  ])


(* CGFILLJMP cgReg, holeReg, htmpLab, holeLab, targReg, ttmpLab, targLab

   The macro is used to fill the hole of an inter-template jump.

   cgReg is a register pointing to a region ref.
   holeReg is a register holding the offset of the copy of the
     template containing the hole to fill.
   htmpLab is the label of the original template with the hole.
   holeLab is the label of the hole in the original template.    
   targReg is a register holding the offset of the copy of the
     template containing the target.
   ttmpLab is the label of the original template with the target.
   targLab is the label of the target in the original template.

   The offset of the hole is calculated as follows:

     holeReg+(holeLab+1-htmpLab-4)=holeReg+(holeLab-htmpLab-3)

   1 is added for the JMP opcode, 4 is subtracted due to template length.

   The target address is calculated as follows:

     (targReg-holeReg)+(targLab-ttmpLab-4)-(holeLab-htmpLab-4+1+4)=
     (targReg-holeReg)+(targLab-ttmpLab)-(holeLab-htmpLab)-5

   4 is subtracted from each template due to template length.  Adding 1
   and 4 are due to the JMP opcode and address. *)


let cgfilljmp_macro cgReg holeReg htmpLab holeLab targReg ttmpLab targLab =
   [  (* Save the offset to the template copy with the target *)
      Push (Reg targReg, []);

      (* Put the fill value into targReg *)
      ArithBin (Sub, Reg targReg, Reg holeReg);
      ArithBin (Add,
                Reg targReg,
		Immed ((targLab -$ ttmpLab)
			 -$ (holeLab -$ htmpLab) -$ (int_to_int32 5)));

      (* Save the offset to the template copy with the hole *)
      Push (Reg holeReg, []);

      (* Get a *pointer* to the template copy with the hole *)
      ArithBin(Add,  
               Reg holeReg,
               Prjr((cgReg,[]),
                    i32_0,
                    None));
      ArithBin(Add,           (* This skips over fills, length, current *)
               Reg holeReg,
               Immed (int_to_int32 12));
   
      (* Fill the hole *)
      Mov (Prjr ((holeReg,[]),(holeLab -$ htmpLab -$ i32_3),None),
	     (Reg targReg,[]));

      (* Restore the offset to the template copy with the hole *)
      Pop (Reg holeReg);

      (* Restore the offset to the template copy with the target *)
      Pop (Reg targReg)
   ]


(*
* CGFILLJCC holeReg, htmplab, holeLab, targReg, ttmpLab, targLab          *
*								          *
* This is the same as CGFILLMP, except a jcc instr filled in.           *

   cgReg is a register pointing to a region ref.
   holeReg is a register holding the offset of the copy of the
     template containing the hole to fill.
   htmpLab is the label of the original template with the hole.
   holeLab is the label of the hole in the original template.    
   targReg is a register holding the offset of the copy of the
     template containing the target.
   ttmpLab is the label of the original template with the target.
   targLab is the label of the target in the original template.

   The offset of the hole is calculated as follows:

     holeReg+(holeLab+2-htmpLab-4)=holeReg+(holeLab-htmpLab-2)

   2 is added for the JCC opcode, 4 is subtracted due to template length.

   The target address is calculated as follows:

     (targReg-holeReg)+(targLab-ttmpLab-4)-(holeLab-htmpLab-4+2+4)=
     (targReg-holeReg)+(targLab-ttmpLab)-(holeLab-htmpLab)-6

   4 is subtracted from each template due to template length.  Adding 2
   and 4 are due to the JCC opcode and address. *)

let cgfilljcc_macro cgReg holeReg htmpLab holeLab targReg ttmpLab targLab =
   [  (* Save the offset to the template copy with the target *)
      Push (Reg targReg, []);

      (* Put the fill value into targReg *)
      ArithBin (Sub, Reg targReg, Reg holeReg);
      ArithBin (Add,
                Reg targReg,
		Immed ((targLab -$ ttmpLab)
			 -$ (holeLab -$ htmpLab) -$ (int_to_int32 6)));

      (* Save the offset to the template copy with the hole *)
      Push (Reg holeReg, []);

      (* Get a *pointer* to the template copy with the hole *)
      ArithBin(Add,  
               Reg holeReg,
               Prjr((cgReg,[]),
                    i32_0,
                    None));
      ArithBin(Add,           (* This skips over fills, length, current *)
               Reg holeReg,
               Immed (int_to_int32 12));
   
      (* Fill the hole *)
      Mov (Prjr ((holeReg,[]),(holeLab -$ htmpLab -$ i32_2),None),
	     (Reg targReg,[]));

      (* Restore the offset to the template copy with the hole *)
      Pop (Reg holeReg);

      (* Restore the offset to the template copy with the target *)
      Pop (Reg targReg)
   ]


(* CGEND reg

   Finish code generation for the region in reg: leave a pointer to
   the resulting function in register reg. *)
let cgend_macro reg =
  preserve_regs (Some reg)
    [ Push (Reg reg, []);
      Mov (Reg Eax, (Addr cg_end_label, []));
      Call (Reg Eax, []);
      ArithBin (Add, (Reg Esp), (Immed i32_4))  ]


(* CGHOLE reg, tmplLab, holeLab			

   Declares a new hole, holeLab, in the template whose first
   instruction has label tmplLab.  The instruction will move the
   contents of the hole into register reg when the hole has been
   filled.

   holeLab is a BINDING OCCURRENCE. *)
(* XXX: eliminate the tmplLab argument, which appears for historical
   reasons *)

let cghole_macro reg tmplLab holeLab =
   (**) (* Bind holeLab to this instruction. *)
   (* XXX: it would be nicer to bind it to the location of the hole *)
   [ Mov (Reg reg, (Immed i32_0,[])) ]


(* CGHOLEJMP tmplLab, holeLab
 
   Same as CGHOLE, except the instruction with a hole is a JMP
   instruction instead of a MOV instruction. *)
(* XXX: eliminate the tmplLab argument, which appears for historical
   reasons *)

let cgholejmp_macro tmplLab holeLab =
   (**)(* Bind holeLab to this instr. *) 
   (* XXX: it would be nicer to bind it to the location of the hole *)
  [ Jmp (Addr external_addr, []) ]


(* CGHOLEJCC cc, tmplLab, holeLab
 
   Same as CGHOLE, except the instruction with a hole is a JCC
   instruction instead of a MOV instruction. *)
(* XXX: eliminate the tmplLab argument, which appears for historical
   reasons *)

let cgholejcc_macro cc tmplLab holeLab =
   (**)(* Bind holeLab to this instr *)
   (* XXX: it would be nicer to bind it to the location of the hole *)
   [ Jcc (cc, (external_addr,[]), None) ]
