(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker, Dan Grossman          *)
(*     August 1999, all rights reserved.                              *)
(**********************************************************************)
(* dasm.ml
 *
 * Disassembles an object file and associated .to file to produce a
 * TAL implementation.
 *)

let pp i = 
  print_string ("Got here " ^ (string_of_int i) ^"\n");
  flush stdout;
  ()
;;


open Tal;;
open Talbin;;
open Talbinin;;
open Numtypes;;
open Disobjfile;;
open Utilities;;

let debug f = 
  let fmt = Format.std_formatter in 
  let o = Talpp.std_options in 
  Format.pp_open_hvbox fmt 0;
  f fmt o;
  Format.pp_print_newline fmt ();
  Format.pp_print_flush fmt (); 
  () 
;;


module type dasm = sig
   val disassemble : bool -> string -> string -> Tal.tal_imp
end 

module DasmFunc (X: sig 
   type in_channel
   val input_char : in_channel -> char
   val input_byte : in_channel -> int
   val pos_in : in_channel -> int 
   val seek_in : in_channel -> int -> unit
   val open_in_bin : string -> in_channel
   val close_in : in_channel -> unit
   val really_input : in_channel -> string -> int -> int -> unit
end ) : dasm = struct

open X
module TalbininX = Talbinin.TalbininFunc(X)
module Discoff = Discoff.DiscoffFunc(X)
module Diself = Diself.DiselfFunc(X)

type binformat = COFF | ELF;;

let objformat =
  match Sys.os_type with
    "Unix" -> ELF
  | "Win32" -> COFF
  | _ -> Printf.eprintf "Unknown operating system\n"; exit 1

let fail objf s = 
  let (sect,addr) = objf.obj_get_pos() in
  Printf.eprintf "dasm (sect=%d, addr=0x%x): %s\n"
    sect addr s;
  raise Gcdfec.Exit;
;;

let get_code_blocks code_annots objf = 
  (* the annotated code blocks -- initialized with the types only *)
  let code_blocks = 
    let bogus_id = Identifier.id_new "\000bogus" in
    Array.init (Array.length code_annots) 
      (fun i -> (bogus_id,fst(code_annots.(i)), Array.of_list [])) in
  (* seek the code in the object file *)
  objf.obj_seek_code();
  let merge_inst_annots = merge_inst_annots objf in
  (* get the code blocks and merge with the annotations *)
  (* i is the index of the current code block *)
  let i = ref 0 in
  (* gets a list of intructions and the next list of labels *)
  let rec get_rest_instrs () = 
    match objf.obj_get_labels 0 with
      [] -> (try get_instrs () with End_of_Section -> ([],[]))
    | labs -> ([],List.rev(labs)) 
  and get_instrs () = 
    let instr = Disasmx86.get_instr objf in
    let (instrs,labs) = get_rest_instrs() in
    (instr::instrs,labs)
  in
  (* sets the code blocks -- we prime the pump with the first set of labels *)
  let rec set_code_blocks labs = 
    match labs with
      (* no labels means we're at the end of the section *)
      [] -> 
	(* Check that there aren't more annotations than code blocks. *)
	if (!i) <> (Array.length code_blocks) then
	  fail objf "more annotations than code blocks!"
	else ()
      (* one label -- should be followed by instructions which we suck
       * in, merge with the annotations, and then bang into the code_blocks
       * array.  As a side-effect of reading the instructions, we get the
       * next list of labels and loop around.  Note that the labels are
       * sucked in in what appears to be reverse order from the original
       * TAL file. *)
    | [lab] -> 
	let (_,copt,_) = code_blocks.(!i) in
	let (raw_instrs,next_labs) = get_instrs() in
	let instrs = merge_inst_annots (snd code_annots.(!i)) raw_instrs in
	code_blocks.(!i) <- (lab,copt,Array.of_list instrs);
	i := (!i) + 1; 
	set_code_blocks next_labs
    | lab::rest -> 
	(* we have more than one label -- so the instructions are empty.
	 * However, the annotations may not be because we may have virtual
	 * instructions, so we have to merge them and then loop around with
	 * the rest of the labels. *)
	let (_,copt,_) = code_blocks.(!i) in
	let instrs = merge_inst_annots (snd code_annots.(!i)) [] in
	code_blocks.(!i) <- (lab,copt,Array.of_list instrs); 
	i := (!i) + 1; 
	set_code_blocks rest in
  (* start the whole thing off by reading in the first list of labels 
   * and skip past the ".text" label. *)
  let init_labs = 
    (match List.rev (objf.obj_get_labels 0) with
      [] -> fail objf "no initial text labels!"
    | t1::rest -> 
	if (Identifier.id_to_string t1) = ".text" then rest
	else fail objf "expecting .text label!") in
  set_code_blocks init_labs;
  code_blocks
;;

let get_data_blocks data_annots objf = 
  objf.obj_seek_data();
  let data_blocks = 
    Array.init (Array.length data_annots)
      (fun i -> (Identifier.id_new "\000bogusd",fst(data_annots.(i)),
		 ([],snd(snd(data_annots.(i)))))) in
  let i = ref 0 in
  let rec set_data_blocks labs = 
    match labs with
      [] -> 
	if (!i) <> (Array.length data_blocks) then
	  fail objf "more annotations than data blocks!"
	else ()
    | [lab] ->
       (* Printf.printf "setting 1 label %s\n" (Identifier.id_to_string lab);*)
	(* FMS: XXX
	   This was crashing when data items of size 0 were being created.
	   I don't know why we were creating data items of size 0.
	   This "if" solves the problem but may not be the right solution.
	   *)
	if (!i) == (Array.length data_blocks) then () else
	let (_,copt,(_,cs)) = data_blocks.(!i) 	in
	let annots = (fst(snd(data_annots.(!i)))) in
	let get_data a = 
	  match a with
	    An_dlabel c -> 
	      (match objf.obj_get_reloc () with
		None -> fail objf "expecting label as data!"
	      |	Some lab -> 
		  let b0 = objf.obj_get_byte() in
		  let b1 = objf.obj_get_byte() in
		  let b2 = objf.obj_get_byte() in
		  let b3 = objf.obj_get_byte() in
		  let v = (b3 lsl 24) lor (b2 lsl 16) lor (b1 lsl 8) lor b0 in
		  if v <> 0 then fail objf "relocation adjustment";
		  Dlabel(lab,c))
	  | An_dbytes (len) ->
	      if len < 0 then fail objf "An_dbytes: size is less than 0"
	      else 
		let s = String.create len in
		for i = 0 to len - 1 do
		  String.set s i (Char.chr (objf.obj_get_byte()))
		done; Dbytes (s)
	  | An_d2bytes ->
	      let lo = objf.obj_get_byte() in
	      let hi = objf.obj_get_byte() in
	      let v = int_to_int32((hi lsl 8) lor lo) in
	      D2bytes(int32_to_int16 v)
	  | An_d4bytes c ->
	      let b0 = objf.obj_get_byte() in
	      let b1 = objf.obj_get_byte() in
	      let b2 = objf.obj_get_byte() in
	      let b3 = objf.obj_get_byte() in
	      let v = (b3 lsl 24) lor (b2 lsl 16) lor (b1 lsl 8) lor b0 in
	      D4bytes(int_to_int32 v,c)
	  | An_dfloat32 ->
	      let b0 = objf.obj_get_byte() in
	      let b1 = objf.obj_get_byte() in
	      let b2 = objf.obj_get_byte() in
	      let b3 = objf.obj_get_byte() in
	      Dfloat32 (bytes_to_float32 [b0;b1;b2;b3])
	  | An_dfloat64 ->
	      let b0 = objf.obj_get_byte() in
	      let b1 = objf.obj_get_byte() in
	      let b2 = objf.obj_get_byte() in
	      let b3 = objf.obj_get_byte() in
	      let b4 = objf.obj_get_byte() in
	      let b5 = objf.obj_get_byte() in
	      let b6 = objf.obj_get_byte() in
	      let b7 = objf.obj_get_byte() in
	      Dfloat64 (bytes_to_float64 [b0;b1;b2;b3;b4;b5;b6;b7])
	  | An_djunk -> 
	      let b0 = objf.obj_get_byte() in
	      let b1 = objf.obj_get_byte() in
	      let b2 = objf.obj_get_byte() in
	      let b3 = objf.obj_get_byte() in
	      let v = (b3 lsl 24) lor (b2 lsl 16) lor (b1 lsl 8) lor b0 in
	      if v <> 0 then fail objf "data supposed to be junk is non-zero";
	      Djunk
	  | An_drep ri -> 
	       (* First four bytes describe the size *)
	       let b0 = objf.obj_get_byte() in
	       let b1 = objf.obj_get_byte() in
	       let b2 = objf.obj_get_byte() in
	       let b3 = objf.obj_get_byte() in
	       let len = (b3 lsl 24) lor (b2 lsl 16) lor (b1 lsl 8) lor b0 in
	       let str = String.create len in 	       
	       for i = 0 to len - 1 do 
		  String.set str i (Char.chr (objf.obj_get_byte()))
		done;
	       let strchan = Stringchan.from_string str in 
	       Drep 
		  ((match ri with 
		     An_con -> RCon (Talbinin.Str.read_in_con strchan)
		   | An_kind -> RKind (Talbinin.Str.read_in_kind strchan)
		   | An_label -> RLabel (Talbinin.Str.read_in_label strchan)),
		     ref (Some (str)))
	  | An_dup -> Dup
	  | An_ddown -> Ddown in
	data_blocks.(!i) <- (lab,copt,(List.map get_data annots,cs));
	objf.obj_align4 ();
	i := (!i) + 1;
	set_data_blocks (List.rev (objf.obj_get_labels 0))	  
    | lab::rest -> 
      (* Printf.printf "setting n labels %s\n" (Identifier.id_to_string lab);*)
	let (_,copt,(_,cs)) = data_blocks.(!i) in
	let annots = (fst(snd(data_annots.(!i)))) in
	let get_data a = 
	  match a with
	    An_dup -> Dup
	  | An_ddown -> Ddown
	  | _ -> fail objf "empty data block but bad annotation" in
	data_blocks.(!i) <- (lab,copt,(List.map get_data annots,cs));
	i := (!i) + 1;
	set_data_blocks rest in
  let init_labs = 
    (match (objf.obj_align4 (); List.rev (objf.obj_get_labels 0)) with
      [] -> fail objf "no initial data labels!"
    | t1::rest -> 
	if (Identifier.id_to_string t1) = ".data" then rest
	else fail objf "expecting .data label!") in
(*    Printf.printf "initial data labels:\n";
    List.iter (fun x -> Printf.printf "%s\n" (Identifier.id_to_string x)) 
    init_labs;
    Printf.printf "--------------------\n"; *)

  set_data_blocks init_labs;
  data_blocks
;;

let disassemble disassem_internals tofile objfile =
  let tofile_ch = open_in_bin tofile in
  Gcdfec.reset_fe "dasm";
  (* read the .to file *)
  let tal_info = TalbininX.read_tal_info disassem_internals tofile_ch in
  let code_annots = tal_info.ti_code_annots in
  let data_annots = tal_info.ti_data_annots in
  
  close_in tofile_ch;
  if disassem_internals then
    begin
      (* open the object file and suck in headers and so forth *)
      let objf = 
	match objformat with
	  COFF -> Discoff.coff_objfile objfile
	| ELF -> Diself.elf_objfile objfile
      in
      (* get the code blocks *)
      let code_blocks = get_code_blocks code_annots objf in
      (* get the data blocks *)
      let data_blocks = get_data_blocks data_annots objf in
      (* get the template blocks -- TO DO *)
      let templates = Array.of_list [] in
      objf.obj_close();
(*      let impl = *)
      { imports = tal_info.ti_imports; 
	exports = tal_info.ti_exports; 
	imp_kindabbrevs = tal_info.ti_kindabbrevs; (* LX *)
	imp_abbrevs = tal_info.ti_abbrevs;
	con_blocks = tal_info.ti_con_blocks; 
	code_blocks = code_blocks;
	data_blocks = data_blocks; 
	templates = templates
      } 
(*      in

      Talout.write_imp "foo" "temp" impl;
      impl *)

    end
  else 
    { imports = tal_info.ti_imports;
      exports = tal_info.ti_exports;
      imp_abbrevs = [||];
      imp_kindabbrevs = [||]; (* LX *)
      con_blocks = [||];
      code_blocks = [||];
      data_blocks = [||];
      templates = [||]
    } 
;;

end (* DasmFunc *)

module Chan = DasmFunc(Pervasives)
module Str = DasmFunc (Stringchan)
