(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker, Dan Grossman          *)
(*     August 1999, all rights reserved.                              *)
(**********************************************************************)
(* talbin.ml 
 *
 * Defines annotations for the binary representation of tal files and
 * operations for splitting out and merging annotations.  
 *)
open Utilities;;
open Numtypes;;
open Identifier;;
open Tal;;

let fail s = 
  (Printf.eprintf "talbin: %s\n" s; flush stderr; raise Gcdfec.Exit)
;;

(* Most of these should be self-explanatory except for the An_opi
 * constructors -- read the source for these to see how they attach
 * themselves to genops.  
 *)
type inst_annot = 
    An_none of int (* no annotation for next i instrs.  1 <= i <= 255 *)
  | An_op1 of coercion list 
  | An_op2 of coercion list * coercion list
  | An_op3 of coercion list * coercion list * coercion list
  | An_op4 of coercion list * coercion list * coercion list * coercion list
  | An_jcc of coercion list * inst_annot list
  | An_coerce of genop * coercion list
  | An_coercename of identifier * coercion list
  | An_fallthru of con list
  | An_malloc of identifier * (mallocarg option)
  | An_proof of (identifier * con list) list
  | An_unpacknomove of identifier * reg * coercion list
  | An_unpackmove of identifier * coercion list * coercion list
  | An_sunpack of identifier * genop
  | An_nameobj of identifier * genop
  | An_forgetunique of identifier
  | An_removename of identifier
(* LX *)
  | An_vcase of int32 * con * identifier * genop coerce 
  | An_letprod of identifier list * con
  | An_letroll of identifier * con 
(* end LX *)
(* Cyclone *)
  | An_cgstart of identifier * con
  | An_cgdump of identifier * identifier
  | An_cghole of identifier * identifier
  | An_cgholejmp of identifier * identifier coerce
  | An_cgholejcc of identifier * identifier coerce
  | An_cgfill of identifier * identifier
  | An_cgfilljmp of identifier * identifier * identifier * identifier
  | An_cgfilljcc of identifier * identifier * identifier * identifier
  | An_cgforget of identifier * identifier
  | An_cgend
(* end Cyclone *)
type arep_item = An_con | An_kind | An_label
type data_annot = 
    An_dlabel of coercion list
  | An_dbytes of int    (* number of bytes *)
  | An_d2bytes
  | An_d4bytes of coercion list
  | An_dfloat32
  | An_dfloat64
  | An_djunk
  | An_dup
  | An_ddown
  | An_drep of arep_item
;;

let print_inst_annot ia = 
  match ia with
    An_none i -> print_string "An_none "; print_int i
  | An_op1 _ -> print_string "An_op1"
  | An_op2 (_,_) -> print_string "An_op2"
  | An_op3 (_,_,_) -> print_string "An_op3"
  | An_op4 (_,_,_,_) -> print_string "An_op4"
  | An_jcc (_,_) -> print_string "An_jcc"
  | An_coerce _ -> print_string "An_coerce"
  | An_coercename _ -> print_string "An_coercename"
  | An_fallthru _ -> print_string "An_fallthru"
  | An_malloc _ -> print_string "An_malloc"
  | An_proof _ -> print_string "An_proof"
  | An_unpacknomove _ -> print_string "An_unpacknomove"
  | An_unpackmove _ -> print_string "An_unpackmove"
  | An_sunpack _ -> print_string "An_sunpack"
  | An_nameobj _ -> print_string "An_nameobj"
  | An_forgetunique _ -> print_string "An_forgetunique"
  | An_removename _ -> print_string "An_removename"
  (* LX *)
  | An_vcase _ -> print_string "An_vcase"
  | An_letprod _ -> print_string "An_letprod"
  | An_letroll _ -> print_string "An_letroll"
(* end LX *)
(* Cyclone *)
  | An_cgstart   _ -> print_string "An_cgstart"
  | An_cgdump    _ -> print_string "An_cgdump"
  | An_cghole    _ -> print_string "An_cghole"
  | An_cgholejmp _ -> print_string "An_cgholejmp"
  | An_cgholejcc _ -> print_string "An_cgholejcc"
  | An_cgfill    _ -> print_string "An_cgfill"
  | An_cgfilljmp _ -> print_string "An_cgfilljmp"
  | An_cgfilljcc _ -> print_string "An_cgfilljcc"
  | An_cgforget  _ -> print_string "An_cgforget"
  | An_cgend       -> print_string "An_cgend"
(* end Cyclone *)

;;

let noannot = An_none 1;;

(* return an annotation for an instruction *)
let rec inst_annot label_map i = 
  let gop_inner g = 
    match g with
      Immed _ -> []
    | Reg _ -> []
    | Addr x -> 
	(match label_map x with
	  -1 -> []
	| x -> [VirtLabel(x)])
    | Prjr ((r,cs),_,_) -> cs
    | Prjl ((x,cs),_,_) -> 
	(match label_map x with
	  -1 -> cs
	| x -> (VirtLabel(x))::cs) in
  let unop gop = 
    match gop_inner gop with
      [] -> noannot
    | cs -> An_op1 cs in
  let cunop gop cs =
    match gop_inner gop,cs with
      [],[] -> noannot
    | [],cs -> An_op1 cs
    | cs1,cs2 -> An_op2(cs1,cs2) in
  let binop gop1 gop2 = 
    match gop_inner gop1, gop_inner gop2 with
      [],[] -> noannot
    | [],cs -> An_op1 cs
    | c1s,c2s -> An_op2(c1s,c2s) in
  let coerces c1 c2 c3 c4 =
    match c1,c2,c3,c4 with
      [],[],[],[] -> noannot
    | [],[],[],cs -> An_op1 cs
    | [],[],cs1,cs2 -> An_op2(cs1,cs2)
    | [],cs1,cs2,cs3 -> An_op3(cs1,cs2,cs3)
    | cs1,cs2,cs3,cs4 -> An_op4(cs1,cs2,cs3,cs4) in
  let cbinop gop1 c1 gop2 c2 = coerces (gop_inner gop1) c1 (gop_inner gop2) c2 
  in
  match i with
    ArithBin(_,gop1,gop2) -> binop gop1 gop2
  | ArithUn(_,gop) -> unop gop
  | ArithMD(_,gop) -> unop gop
  | ArithSR(_,gop,_) -> unop gop
  | Bswap _ -> noannot
  | Call (g,cs) -> cunop g cs
  | Clc -> noannot
  | Cmc -> noannot
  | Cmovcc (_,_,(g,c)) -> cunop g c
  | Cmp((g1,c1),(g2,c2)) -> cbinop g1 c1 g2 c2
  | Conv _ -> noannot
  | Imul3(_,g,_) -> unop g
  | Int _ -> noannot
  | Into -> noannot
  | Jcc(_,(x,c),iopt) -> 
      let annots = 
	match iopt with 
	  None -> [] 
	| Some is -> List.map (inst_annot label_map) is in
      (match c,annots,label_map x with
	[],[],-1 -> noannot 
      |	_,_,-1 -> An_jcc(c,annots)
      |	_,_,x -> An_jcc(VirtLabel(x)::c,annots))
  | Jecxz((x,c),iopt) ->
      let annots = 
	match iopt with 
	  None -> [] 
	| Some is -> List.map (inst_annot label_map) is in
      (match c,annots,label_map x with
	[],[],-1 -> noannot 
      |	_,_,-1 -> An_jcc(c,annots)
      |	_,_,x -> An_jcc(VirtLabel(x)::c,annots))
  | Jmp(g,c) -> cunop g c
  | Lahf -> noannot
  | Lea(_,g) -> unop g
  | Loopd((x,c),_) -> 
      (match c,label_map x with 
	[],0 -> noannot 
      |	_,0 -> An_op1 c
      |	_,x -> An_op1 (VirtLabel(x)::c))
    (* JGM:  note that the first list of coercions is empty guaranteeing
       * that we cannot get an An_op4 annotation. *)
  | Mov(g1,(g2,c)) -> coerces [] (gop_inner g1) (gop_inner g2) c
  | Movpart(_,g1,_,g2,_) -> binop g1 g2
  | Nop -> noannot
  | Pop g -> unop g
  | Popad -> noannot
  | Popfd -> noannot
  | Push(g,c) -> cunop g c
  | Pushad -> noannot
  | Pushfd -> noannot
  | Retn _ -> noannot
  | Sahf -> noannot
  | Setcc(_,g) -> unop g
  | Shld(g,_,_) -> unop g
  | Shrd(g,_,_) -> unop g
  | Stc -> noannot
  | Test(g1,g2) -> binop g1 g2
  | Xchg(g,_) -> unop g
  | Coerce(g,c) -> An_coerce(g,c)
  | CoerceName(x,c) -> An_coercename(x,c)
  | Comment _ -> noannot
  | Fallthru(cs) -> An_fallthru(cs)
  | Malloc(x,_,mopt) -> An_malloc(x,mopt)
  | Proof xs -> An_proof xs
  | Unpack(x,r,(gop,c)) ->
      begin
	let a = An_unpackmove(x,gop_inner gop,c) in
	match gop with
	  Reg r' -> 
	    if r' = r then An_unpacknomove(x,r,c) else a
	| _ -> a
      end
  | Sunpack(i,g) -> An_sunpack(i,g)
  | Nameobj(x,g) -> An_nameobj(x,g)
  | ForgetUnique x -> An_forgetunique x
  | RemoveName x -> An_removename x
(* Floating Point *)
  | FPnoargs _ -> noannot
  | FPsomeargs (_,args) -> 
      (match args with
	FPstack _ | FPstack2 _ -> noannot
      |	FPgenop (_,g) -> unop g)
(* LX *)	  
  | Vcase (i,c,id,gc) -> An_vcase (i,c,id,gc)
  | Letroll (i,c) -> An_letroll (i,c)
  | Letprod (l,c) -> An_letprod(l,c)
(* end LX *)
(* Cyclone *)
  | CgStart(i,c)                    -> An_cgstart(i,c)
  | CgDump(r1,i1,r2,i2)             -> An_cgdump(i1,i2)
  | CgHole(r,i1,i2)                 -> An_cghole(i1,i2)
  | CgHoleJmp(i1,ic)                -> An_cgholejmp(i1,ic)
  | CgHoleJcc(c,i1,ic)              -> An_cgholejcc(i1,ic)
  | CgFill(r0,r1,i1,i2,r2)          -> An_cgfill(i1,i2)
  | CgFillJmp(r0,r1,i1,i2,r2,i3,i4) -> An_cgfilljmp(i1,i2,i3,i4)
  | CgFillJcc(r0,r1,i1,i2,r2,i3,i4) -> An_cgfilljcc(i1,i2,i3,i4)
  | CgForget(i1,i2)                 -> An_cgforget(i1,i2)
  | CgEnd _                         -> An_cgend
(* end Cyclone *)
(*  | _ -> fail "unimplemented instruction" *)
;;

(* given a list of instruction annotations and a list of instructions,
 * merge them to produce a list of annotated instructions.  *)
let merge_inst_annots objf annots instrs = 
  let labelmap = objf.Disobjfile.obj_get_shared_labels in
  (* replace a label with the intended one by merging away the VirtLabel
     coercion *)
  let merge_virtlabel a i =
    try List.nth (labelmap a) i
    with Dict.Absent -> 
(*      print_string("\n\t");
      print_string(id_to_string a);
*)
      fail "VirtAddress on non-shared address?"
    | _ ->
(*      List.iter (fun id -> print_string(id_to_string id)) (labelmap a);
      print_string " ";
      print_string (id_to_string a); print_string(" "); print_int i;
*)
      fail "merge_virtlabel: index too large"
  in

  (* merge a coercion onto the inner operand of a memory genop *)
  let merge_inner g cs = 
    match cs,g with
      [],_ -> g
    | [VirtLabel(x)],Addr z -> Addr (merge_virtlabel z x)
    | (VirtLabel(x)::cs),(Prjl((z,_),i,sopt)) -> 
	Prjl((merge_virtlabel z x, cs),i,sopt)
    | _,Prjl((x,_),i,sopt) -> Prjl((x,cs),i,sopt)
    | _,Prjr((r,_),i,sopt) -> Prjr((r,cs),i,sopt)
    | _,Immed _ -> fail "merge_inner: Immed"
    | _,Reg   _ -> fail "merge_inner: Reg"
    | _,Addr  _ -> fail "merge_inner: Addr"
  in

  (* merge a single coercion with an instruction -- this is always the
   * right most, inner most genop *)    
  let merge_op1 instr cs = 
    match instr with
      ArithBin(ab,gop1,gop2) -> ArithBin(ab,gop1,merge_inner gop2 cs)
    | ArithUn(u,gop) -> ArithUn(u,merge_inner gop cs)
    | ArithMD(m,gop) -> ArithMD(m,merge_inner gop cs)
    | ArithSR(a,gop,x) -> ArithSR(a,merge_inner gop cs,x)
    | Call (g,_) -> Call(g,cs)
    | Cmovcc(x,y,(g,_)) -> Cmovcc(x,y,(g,cs))
    | Cmp(cg1,(g2,_)) -> Cmp(cg1,(g2,cs))
    | Imul3(x,g,y) -> Imul3(x,merge_inner g cs,y)
    | Jmp(g,_) -> Jmp(g,cs)
    | Lea(x,g) -> Lea(x,merge_inner g cs)
    | Loopd((x,_),w) -> 
      	(match cs with
	  VirtLabel(y)::cs -> 
	    Loopd((merge_virtlabel x y, cs),w)
      	| _ -> Loopd((x,cs),w))
    | Mov(g1,(g2,_)) -> Mov(g1,(g2,cs))
    | Movpart(x,g1,y,g2,z) -> Movpart(x,g1,y,merge_inner g2 cs,z)
    | Pop g -> Pop(merge_inner g cs)
    | Push(g,_) -> Push(g,cs)
    | Setcc(x,g) -> Setcc(x,merge_inner g cs)
    | Shld(g,x,y) -> Shld(merge_inner g cs,x,y)
    | Shrd(g,x,y) -> Shrd(merge_inner g cs,x,y)
    | Test(g1,g2) -> Test(g1,merge_inner g2 cs)
    | Xchg(g,x) -> Xchg(merge_inner g cs,x)
    | i -> 
      	(Talpp.print_instruction Format.std_formatter Talpp.std_options i;
	 Format.print_newline(); fail "merge_op1: bad instruction")
  in

  (* merge two coercions with an instruction -- either we have two genops
   * with inner coercions [binop] (isn't this impossible?) or we have a 
   * single coerced genop with an inner coercion [cunop] or we have
   * two genops with the right-most genop having both an inner and outer
   * coercion.
   *)
  let merge_op2 instr cs1 cs2 = 
    match instr with
      ArithBin(a,gop1,gop2) -> 
      	ArithBin(a,merge_inner gop1 cs1,merge_inner gop2 cs2)
    | Call(g,_) -> Call(merge_inner g cs1,cs2)
    | Cmovcc(x,y,(g,_)) -> Cmovcc(x,y,(merge_inner g cs1,cs2))
    | Cmp(cg1,(g2,_)) -> Cmp(cg1,(merge_inner g2 cs1,cs2))
    | Jmp(g,_) -> Jmp(merge_inner g cs1,cs2)
    | Mov(g1,(g2,_)) -> Mov(g1,(merge_inner g2 cs1,cs2))
    | Movpart(x,g1,y,g2,z) -> 
      	Movpart(x,merge_inner g1 cs1,y,merge_inner g2 cs2,z)
    | Push(g,_) -> Push(merge_inner g cs1,cs2)
    | Test(g1,g2) -> Test(merge_inner g1 cs1,merge_inner g2 cs2)
    | i -> 
      	(Talpp.print_instruction Format.std_formatter Talpp.std_options i;
	 Format.print_newline(); fail "merge_op2 bad instruction")
  in
  
  (* merge three coercions with an instruction -- this has to be a cbinop
   * so it must be Cmp or Mov. *)
  let merge_op3 instr cs1 cs2 cs3 = 
    match instr with
      Cmp((g1,_),(g2,_)) -> Cmp((g1,cs1),(merge_inner g2 cs2,cs3))
    | Mov(g1,(g2,_)) -> Mov(merge_inner g1 cs1,(merge_inner g2 cs2,cs3))
    | i ->  
      	(Talpp.print_instruction Format.std_formatter Talpp.std_options i;
	 Format.print_newline(); fail "merge_op3 bad instruction")
  in

  (* only cmp can generate this and only when the left-operand is a memory
   * operand with two coercions (guaranteeing that there's no innder coercion
   * for the right operand, but we don't encode this.)
   *)
  let merge_op4 instr cs1 cs2 cs3 cs4 = 
    match instr with
      Cmp((g1,_),(g2,_)) -> Cmp((merge_inner g1 cs1,cs2),
			      	(merge_inner g2 cs3,cs4))
    | _ -> 
      	(Talpp.print_instruction Format.std_formatter Talpp.std_options instr;
	 Format.print_newline(); fail "merge_op4 bad instruction")
  in

  let rec merge annots is accum =
(*
    if annots <> [] then (print_inst_annot (List.hd annots); 
			  print_string "\n") else ();
*)
    match annots, is with
      [],[] -> List.rev accum
    | (An_none n)::rest,is -> 
	if n < 0 then fail ("An_none "^string_of_int(n)) else
	let rec loop n is accum = 
	  match n,is with
	    0,_ -> merge rest is accum
	  | _,hd::tl -> loop (n-1) tl (hd::accum)
	  | _,[] -> fail ("An_none length mismatch")
	in loop n is accum
    | (An_op1(cs))::rest,i::is -> 
	merge rest is ((merge_op1 i cs)::accum)
    | (An_op2(cs1,cs2))::rest,i::is ->
	merge rest is ((merge_op2 i cs1 cs2)::accum)
    | (An_op3(cs1,cs2,cs3))::rest,i::is ->
	merge rest is ((merge_op3 i cs1 cs2 cs3)::accum)
    | (An_op4(cs1,cs2,cs3,cs4))::rest,i::is ->
	merge rest is ((merge_op4 i cs1 cs2 cs3 cs4)::accum)
    | (An_jcc(cs,annots2))::annots,(Jcc(cc,(x,_),_))::is ->
	let (id,cs) = 
	  (match cs with
	    (VirtLabel(y)::cs) -> (merge_virtlabel x y, cs)
	  | _ -> (x,cs)) in
	(match merge annots2 [] [] with
	  [] -> merge annots is ((Jcc(cc,(id,cs),None))::accum)
	| is2 -> merge annots is ((Jcc(cc,(id,cs),Some is2))::accum))
    | (An_jcc(cs,annots2))::annots,(Jecxz((x,_),_))::is ->
	let (id,cs) = 
	  (match cs with
	    (VirtLabel(y)::cs) -> (merge_virtlabel x y, cs)
	  | _ -> (x,cs)) in
	(match merge annots2 [] [] with
	  [] -> merge annots is ((Jecxz((id,cs),None))::accum)
	| is2 -> merge annots is ((Jecxz((id,cs),Some is2))::accum))
    | (An_coerce(g,cs))::rest,_ -> merge rest is (Coerce(g,cs)::accum)
    | (An_coercename(x,cs))::rest,_ ->
	merge rest is (CoerceName(x,cs)::accum)
    | (An_fallthru(cs))::rest,_ -> merge rest is (Fallthru(cs)::accum)
    | (An_malloc(x,mopt))::annots,
	(* NB:  this must match the malloc macro in talasmx86.ml *)
	(Push (Immed size,[]))::
	(Mov (Reg Eax, (Addr malloc_label, [])))::
	(Call (Reg Eax,[]))::
	(ArithBin (Add, Reg Esp,Immed i32))::
	(Test (Reg Eax, Reg Eax))::
	(Jcc (Eq, (out_of_memory_error_label, []), None))::insts ->
	  if id_compare malloc_label (id_of_string "_GC_malloc") = 0 &&
	     id_compare out_of_memory_error_label 
	      (id_of_string "_out_of_memory") = 0 &&
	    i32 = i32_4 then
	    merge annots insts ((Malloc(x,size,mopt))::accum)
	  else fail "merge: malloc -- bad instructions"
    | (An_proof(xs))::rest,_ -> merge rest is (Proof(xs)::accum)
    | (An_unpacknomove(x,r,cs))::rest,_ -> 
	merge rest is (Unpack(x,r,(Reg r,cs))::accum)
    | (An_unpackmove(x,cs1,cs2))::rest, (Mov (Reg r,(gop,_)))::insts ->
	merge rest insts ((Unpack(x,r,(merge_inner gop cs1,cs2)))::accum)
    | (An_sunpack(i,g))::rest,_    -> merge rest is (Sunpack(i,g)::accum)
    | (An_nameobj(x,g))::rest,_    -> merge rest is (Nameobj(x,g)::accum)
    | (An_forgetunique(x))::rest,_ -> merge rest is (ForgetUnique(x)::accum)
    | (An_removename(x))::rest,_   -> merge rest is (RemoveName(x)::accum)
(* LX *)
    | (An_vcase (i,c,id,gc))::rest,_ -> merge rest is (Vcase (i,c,id,gc)::accum)
    | (An_letprod (il,c))::rest,_    -> merge rest is (Letprod(il,c)::accum)
    | (An_letroll (i,c))::rest,_     -> merge rest is (Letroll (i,c)::accum)
(* end LX *)
(* CYCLONE *)
    (* These patterns should match the macros defined in cyclone.ml *)
    | (An_cgstart(i,c))::rest,
        (Push (Reg Ebx, [])):: (* push3regs Eax*)
	(Push (Reg Ecx, []))::
	(Push (Reg Edx, []))::
        (Mov (Reg Eax, (Addr cg_start_label, [])))::
        (Call (Reg Eax, []))::
	(Pop (Reg Edx))::      (* pop3regs Eax*)
	(Pop (Reg Ecx))::
	(Pop (Reg Ebx))::
        insts ->
	  if id_compare cg_start_label Cyclone.cg_start_label = 0
          then
	    merge rest insts ((CgStart(i,c))::accum)
	  else
            fail "merge: cgstart -- bad instructions"
    | (An_cgdump(i1,i2))::rest,
        (Push (Reg Ebx, [])):: (* push3regs Eax *)
	(Push (Reg Ecx, []))::
	(Push (Reg Edx, []))::
        (Push (Addr lab, []))::
        (Push (Reg cgReg,[]))::
        (Mov (Reg Eax, (Addr cg_dump_label, [])))::
        (Call (Reg Eax, []))::
        (ArithBin (Add, (Reg Esp), (Immed i32)))::
	(Pop (Reg Edx))::      (* pop3regs Eax *)
	(Pop (Reg Ecx))::
	(Pop (Reg Ebx))::
        insts ->
          if id_compare cg_dump_label Cyclone.cg_dump_label <> 0
              or id_compare lab i2 <> 0
              or i32 <> i32_8
          then 
            fail "merge: cgdump -- bad instructions 1"
	  else
	    merge rest insts ((CgDump(cgReg,i1,Eax,i2))::accum)
    | (An_cgdump(i1,i2))::rest,
        (Push (Reg r1, [])):: (* push3regs reg<>Eax *)
	(Push (Reg r2, []))::
	(Push (Reg r3, []))::
        (Push (Addr lab, []))::
        (Push (Reg cgReg,[]))::
        (Mov (Reg Eax, (Addr cg_dump_label, [])))::
        (Call (Reg Eax, []))::
        (ArithBin (Add, (Reg Esp), (Immed i32)))::
        (Mov (Reg reg, (Reg Eax, [])))::
        (Pop (Reg r3')):: (* pop3regs reg *)
	(Pop (Reg r2'))::
	(Pop (Reg r1'))::
        insts ->
          if id_compare cg_dump_label Cyclone.cg_dump_label <> 0
              or id_compare lab i2 <> 0
              or i32 <> i32_8
              or r1<>r1' or r2<>r2' or r3<>r3'
              or (reg=Ebx && (r1<>Eax or r2<>Ecx or r3<>Edx))
              or (reg=Ecx && (r1<>Eax or r2<>Ebx or r3<>Edx))
              or (reg=Edx && (r1<>Eax or r2<>Ebx or r3<>Ecx))
              or (reg<>Ebx && reg<>Ecx && reg<>Edx)
          then
            fail "merge: cgdump -- bad instructions 2"
          else
	    merge rest insts ((CgDump(cgReg,i1,reg,i2))::accum)
    | (An_cghole(i1,i2))::rest,_ ->
        fail "unimplemented RTCG annotation An_cghole"
    | (An_cgholejmp(i1,ic))::rest,_ ->
        fail "unimplemented RTCG annotation An_cgholejmp"
    | (An_cgholejcc(i1,ic))::rest,_ ->
        fail "unimplemented RTCG annotation An_cgholejcc"
    | (An_cgfill(i1,i2))::rest,
        (Push (Reg Eax, []))::
        (Push (Reg Ebx, []))::
        (Push (Reg Ecx, []))::
        (Push (Reg Edx, []))::
        (ArithBin (Add, (Reg reg1), (Immed disp)))::
        (Push (Reg reg1', []))::
        (Push (Reg reg2, []))::
        (Push (Reg cgReg,[]))::
        (Mov (Reg Eax, (Addr cg_mark_label, [])))::
        (Call (Reg Eax, []))::
        (ArithBin (Add, (Reg Esp), (Immed i32)))::
        (Pop (Reg Edx))::
        (Pop (Reg Ecx))::
        (Pop (Reg Ebx))::
        (Pop (Reg Eax))::
        insts ->
          if id_compare cg_mark_label Cyclone.cg_mark_label <> 0
              or i32 <> (int_to_int32 12)
              or reg1<>reg1'
          then
            fail "merge: cgfill -- bad instructions"
          else
            (try
              let (sec1,off1) =
                Dict.lookup objf.Disobjfile.obj_label_offset i1 in
              let (sec2,off2) =
                Dict.lookup objf.Disobjfile.obj_label_offset i2 in
              let off1 = int_to_int32 off1 in
              let off2 = int_to_int32 off2 in
              if sec1<>sec2
              then
                fail "merge: cgfill -- bad instructions 1"
              else if disp <> (off2 -$ off1 -$ i32_3)
              then
                fail "merge: cgfill -- bad instructions 2"
              else
	        merge rest insts ((CgFill(cgReg,reg1,i1,i2,reg2))::accum)
            with Dict.Absent ->
              Printf.eprintf "Error looking up identifiers:\n";
              Printf.eprintf "%s or %s\n" (id_to_string i1) (id_to_string i2);
              Printf.eprintf "Dictionary entries are:\n";
              Dict.app_dict
                (fun a b -> Printf.eprintf "%s\n" (id_to_string a))
                objf.Disobjfile.obj_label_offset;                
              flush stderr;
              fail "merge: cgfill -- bad instructions 3")
    | (An_cgfilljmp(i1,i2,i3,i4))::rest,_ ->
        fail "unimplemented RTCG annotation An_cgfilljmp"
    | (An_cgfilljcc(i1,i2,i3,i4))::rest,_ ->
        fail "unimplemented RTCG annotation An_cgfilljcc"
    | (An_cgforget(i1,i2))::rest,_ ->
	merge rest is ((CgForget(i1,i2))::accum)
    | (An_cgend)::rest,
        (Push (Reg Ebx, [])):: (* push3regs Eax *)
	(Push (Reg Ecx, []))::
	(Push (Reg Edx, []))::
        (Push (Reg Eax, []))::
        (Mov (Reg Eax, (Addr cg_end_label, [])))::
        (Call (Reg Eax, []))::
        (ArithBin (Add, (Reg Esp), (Immed i32)))::
	(Pop (Reg Edx))::      (* pop3regs Eax *)
	(Pop (Reg Ecx))::
	(Pop (Reg Ebx))::
        insts ->
          if id_compare cg_end_label Cyclone.cg_end_label <> 0
              or i32 <> i32_4
          then 
            fail "merge: cgend -- bad instructions 1"
	  else
	    merge rest insts ((CgEnd Eax)::accum)
    | (An_cgend)::rest,
        (Push (Reg r1, [])):: (* push3regs reg *)
	(Push (Reg r2, []))::
	(Push (Reg r3, []))::
        (Push (Reg reg, []))::
        (Mov (Reg Eax, (Addr cg_end_label, [])))::
        (Call (Reg Eax, []))::
        (ArithBin (Add, (Reg Esp), (Immed i32)))::
        (Mov (Reg reg', (Reg Eax,[])))::
	(Pop (Reg r3'))::      (* pop3regs reg *)
	(Pop (Reg r2'))::
	(Pop (Reg r1'))::
        insts ->
          if id_compare cg_end_label Cyclone.cg_end_label <> 0
              or i32 <> i32_4
              or reg<>reg'
              or r1<>r1' or r2<>r2' or r3<>r3'
              or (reg=Ebx && (r1<>Eax or r2<>Ecx or r3<>Edx))
              or (reg=Ecx && (r1<>Eax or r2<>Ebx or r3<>Edx))
              or (reg=Edx && (r1<>Eax or r2<>Ebx or r3<>Ecx))
              or (reg<>Ebx && reg<>Ecx && reg<>Edx)
          then 
            fail "merge: cgend -- bad instructions 2"
	  else
	    merge rest insts ((CgEnd Eax)::accum)
(* end CYCLONE *)
    | [],_::_ -> fail "merge:too many instructions"
    | _::_,[] -> fail "merge:too many annotations"
    | a::_,i::_ -> 
	print_string "\nAnnotation: "; print_inst_annot a; print_string "\n";
	print_string "Instruction: ";
	Talpp.print_instruction Format.std_formatter Talpp.std_options i;
	Format.print_newline(); 
	fail "merge_inst_annots -- bad instruction" in 
(*
  List.iter (fun a -> (print_inst_annot a; print_string "\n")) annots;
  List.iter (fun i -> (Talpp.print_instruction Format.std_formatter
			 Talpp.std_options i; Format.print_newline ())) instrs;
*)
  merge annots instrs []
;;

let print_data_annot d = 
  match d with
    An_dlabel _ -> print_string "An_dlabel"
  | An_dbytes _ -> print_string "An_dbytes"
  | An_d2bytes  -> print_string "An_d2bytes"
  | An_d4bytes _ -> print_string "An_d4bytes"
  | An_dfloat32 -> print_string "An_dfloat32"
  | An_dfloat64 -> print_string "An_dfloat64"
  | An_djunk -> print_string "An_djunk"
  | An_drep _ -> print_string "An_drep"
  | An_dup -> print_string "An_dup"
  | An_ddown -> print_string "An_ddown"
;;

(* return an annotation for a hunk of data *)
let data_annots ds = 
  let dannot d = 
    match d with
      Dlabel (x,c) -> An_dlabel c
    | Dbytes (s) -> An_dbytes(String.length s)
    | D2bytes _ -> An_d2bytes
    | D4bytes (_,c) -> An_d4bytes c
    | Dfloat32 _ -> An_dfloat32
    | Dfloat64 _ -> An_dfloat64
    | Djunk -> An_djunk
    | Drep (ri,s) -> An_drep (match ri with 
	 RCon _ -> An_con
       | RKind _ -> An_kind
       | RLabel _ -> An_label)
    | Dup -> An_dup
    | Ddown -> An_ddown
  in List.map dannot ds
;;

(* find labels that have no real instructions since they will have the
 * same address as other labels.  All other such labels will need
 * a VirtLabel coercion on their uses
 *)
let virtual_label_map cbv = 
   let has_real_instruction i = 
     let (_,_,insts) = cbv.(i) in
     let len = Array.length insts in
     let rec loop i = 
       if i < len then
	 (not (is_virtual_instruction (insts.(i)))) or (loop (i+1))
       else false in 
     loop 0 
   in
   let has_real_v = Array.init (Array.length cbv) has_real_instruction in
   let d = ref (Dict.empty id_compare) in
   let rec loop_outer i =
     if i = (Array.length cbv)
     then ()
     else 
       begin
	 let rec loop_inner pos =
	   (if i+pos = (Array.length cbv)
	   then fail "virtual block ends code segment");
	   let l,_,_ = cbv.(i+pos) in
	   d := Dict.insert (!d) l pos;
	   if has_real_v.(i+pos)
	   then loop_outer (i+pos+1)
	   else loop_inner (pos+1)
	 in
	 if has_real_v.(i)
	 then loop_outer(i+1)
	 else loop_inner 0
       end
   in
   loop_outer 0;

   let d = (!d) in
   let f x = try Dict.lookup d x with Dict.Absent -> -1 in
   f
;;
