(* Add extra indirection for contents of the type rep *)
let tuple_rep = ref true

(* Given an interface file, this program generates C code for the init 
   function registering the value labels present in that interface file.
   Code is emitted to standard out. *)

(* Keeps track of specialized string definitions *)
let lengths_defined = ref [];;
let add_length l =
  let svar = "rep"^(string_of_int l) in
  if (List.mem l !lengths_defined) then
    (svar,"")
  else
    (lengths_defined := l::(!lengths_defined);
     (svar,
      Printf.sprintf "typedef struct { int len; char s[%d]; } %s;\n" l svar))

(* Generate a string that resembles a C static declaration for
   the provided buffer *)
let make_static_decl varname str ofs len =
  let toString c =
    let cc = Char.code c in
    Printf.sprintf "%d" cc in
  let char_list =
    let l = ref [] in
    for i = ofs to len-1 do
      l := (toString str.[i])::!l;
    done;
    List.rev !l in
  let data = String.concat "," char_list in
  let var = Printf.sprintf "ctx_%s" varname in
  let (svar,svar_decl) = add_length len in
  let decl =
    Printf.sprintf
      "%sstatic %s %s = { %d, { %s }};" svar_decl svar var len data in
  (var,decl)
;;

(* Generate a string that resembles a C static declaration for
   the provided buffer
let make_static_decl varname str ofs len =
  let toString c =
    let cc = Char.code c in
    Printf.sprintf "%d" cc in
  let char_list =
    let l = ref [] in
    for i = ofs to len-1 do
      l := (toString str.[i])::!l;
    done;
    List.rev !l in
  let data = String.concat "," char_list in
  let var = Printf.sprintf "ctx_%s" varname in
  let decl = 
    Printf.sprintf
      "static char chars_%s[] = { %s };\nstatic struct str_internal %s = { %d, chars_%s };" 
      var data var len var in
  (var,decl)
;;
*)

let rec normalize_con c ctxt =
  Talcon.normalize ctxt (Talcon.expand_abbrevs ctxt c)
(*
  let c' = Talcon.normalize ctxt (Talcon.expand_abbrevs ctxt c) in
  (* check if no more expansion needed *)
  let finished = 
    try
      Talcon.eqcon (Talctxt.empty_ctxt) c c';
      true
    with _ -> false in
  (* either done or, iterate again *)
  if finished then c else normalize_con c' ctxt
*)

(* Determines whether the given constructor represents a Popcorn function *)
let is_fun_con c ctxt =
  let c = normalize_con c ctxt in
  let rec skip_forall c i =
    match c with
      { Tal.rcon = rc } ->
	(match rc with
	  Tal.Cforall (id, k, c) -> skip_forall c true
	| _ -> (c,i))
  in
  let isCode c =
    match c with
      { Tal.rcon = rc } -> 
	(match rc with
	  Tal.Ccode _ -> true
	| _ -> false)
  in
(*
  Printf.printf "Checking con:\n";
  Talpp.print_con (Format.std_formatter) Talpp.std_options c;
  Format.pp_print_newline (Format.std_formatter) ();
*)
  let (c,skippedForall) = skip_forall c false in
(*
  Printf.printf "Skipped forall: %b, next con =\n" skippedForall;
  Talpp.print_con (Format.std_formatter) Talpp.std_options c;
  Format.pp_print_newline (Format.std_formatter) ();  
*)
  skippedForall && (isCode c)

let main() =
  try
    let argv = Sys.argv in
    if (Array.length argv) != 2 then
      (Printf.fprintf stderr "usage: %s <tali file>\n" argv.(0);
       exit 1);
    
    if not (Sys.file_exists argv.(1)) then
      (Printf.fprintf stderr "%s: file not found\n" argv.(1);
       exit 1);
    
    let outc = Pervasives.stdout in
    let tali = Talbe.read_int argv.(1) in
    let intname = 
      Filename.chop_extension (Filename.basename argv.(1)) in
    let ctxt = 
      Array.fold_left
        (fun ctxt (id,con) -> 
          (* XXX call check rather than expand_abbrev *)
	  Talctxt.add_abbrev ctxt id (Talcon.expand_abbrevs ctxt con))
	Talctxt.empty_ctxt tali.Tal.int_abbrevs in
(*
      Array.fold_left 
	(fun ctxt (id,con) -> Talctxt.add_abbrev ctxt id con)
	Talctxt.empty_ctxt tali.Tal.int_abbrevs in
*)
    let buf = Buffer.create 16 in

    let indirect_con con =
      Tal.cprod_b [Tal.cfield con Tal.ReadWrite] in

(*
      (Tal.chptr [] 
	 (Some (Tal.cprod [Tal.cfield con Tal.ReadWrite])) 
	 None in
*)    
    (* For each value label, add a type representation *)
    let val_cons =
      Array.map 
	(function (id,con) ->
	  (* for non-functions, need to indirect the constructor *)
	  let (con,indirect) =
	    if not (is_fun_con con ctxt) then
              (con,true)
            else
              (con,false) in
	  (* expand type abbreviations *)
	  let con = normalize_con con ctxt in
          let con = if !tuple_rep then (indirect_con con) else con in
(*
	  Printf.fprintf Pervasives.stderr "after normalizing:\n";
	  Talpp.print_con (Format.err_formatter) Talpp.std_options con;
	  Format.pp_print_newline (Format.err_formatter) ();	  
*)
	  (* generate the type rep *)
	  Buffer.clear buf;
	  Talbinout.Buf.emit_out_con buf con;
	  (* remove leading _ from id *)
	  let idstrname = 
	    let idstr = Identifier.id_to_string id in
	    if idstr.[0] = '_' then
	      String.sub idstr 1 ((String.length idstr) - 1) 
	    else
	      idstr in
	  (idstrname,indirect,Buffer.contents buf))
	tali.Tal.int_vals in

    (* Now output the results *)
    Printf.fprintf outc
      "/*** CODE BELOW GENERATED AUTOMATICALLY BY genCinit.exe ****/\n\n";
    Printf.fprintf outc 
      "typedef struct str_internal {int size; char *chars;} *string;\n";
    Printf.fprintf outc 
      "typedef struct rep_internal {int size; char chars[1];} *rep;\n";

    (* First output the type representations as static data *)
    let rep_vars =
      Array.map
	(function (idstr,indirect,rep) ->
	  let (idrep,decl) = 
	    make_static_decl idstr rep 0 (String.length rep) in
	  Printf.fprintf outc "\n\n/*** %s ***/\n" idstr;
	  Printf.fprintf outc "%s\n" decl;
	  let idlen = String.length idstr in
	  Printf.fprintf outc
	    "static struct str_internal str_%s = { %d, \"%s\" };" 
	    idstr idlen idstr;
	  (idstr,indirect,idrep))
	val_cons in

    (* Now create the init function, registering each of the labels *)
    (* We must not use any registers---the popcorn update function will 
       not properly save callee-saves registers  *)
    Printf.fprintf outc "\nstatic void (*add_sym)(string sym,rep r,void *p);\n";
    Printf.fprintf outc "void dyninit_%s(void *unused,void (*f)(string sym,rep r,void *p), void *unused2) {\n  add_sym = f;\n" intname;
    
    Array.iter
      (function (idstr,indirect,idrep) ->
	Printf.fprintf outc 
	  "  add_sym((string)&str_%s,(rep)&%s,(void *)%c%s);\n" 
	  idstr idrep (if indirect then '&' else ' ') idstr)
      rep_vars;

    Printf.fprintf outc "}\n"

  with e ->
    Printf.printf "Failed with exception %s\n" (Printexc.to_string e)
;;

let _ = main()
;;
      
