(**********************************************************************)
(* (c) Michael Hicks                                                  *)
(*     April 2000, all rights reserved.                               *)
(**********************************************************************)

(* The translation for dynamically loadable files consists of a number
   of transformations to the source file.  In essence, all "externs" are
   translated into members of a single structure global to the file, 
   called the global offset table (GOT).  Accesses to externs are changed
   to be indirected through the GOT.  The GOT is filled in at load-time
   by an added "init" function.  This function looks up the missing
   symbols using functions provided by the running program, and fills
   the GOT appropriately.  Any symbols that should be made available to
   future loaded code are exported using the passed functions.

   The code implements this translation as follows:

    - generate import table
      - slots and default entries
    - change references to externals to be through import table
    - generate init function
      - exports
      - imports
*)

(* NOTES

   I need to rewrite some of this stuff to make it more clear.  Here
   are some invariants I should exploit for carefully:

   - the range of flags.add_varmap is a subset of the range of
     flags.upd_varmap.  upd_varmap consists of the list of functions
     to be added/replaced, while the add_varmap is the list of
     functions that have stub versions (which are themselves in the
     upd_map).  To know if a function is going to be updated, then we
     just need to check the range of the upd_varmap.

   - the implementation code should never refer to the old version
     directly.  The stub code may refer to the old version, but it
     should do so directly (not with an extra level of indirection
     in the update case).

   - we define four sets of variables
     - those to look up immediately, includes 
       - UnresolvedSymbol (for default functions)
       - any of those that we are looking up that are also in the
         range of the update/add varmap (may just be the add varmap).
         Need to look them up before we replace the old value
         in the table.  This also implies that the entry should
         not be indirected through the table.
     - those to update/add
       - on entry (flags.upd_varmap), this list is those that
         will replace existing functions.  During transformation we
         also add variables that are to be exported in addition.
     - those to add
       - these are ones that stub existed for.  Therefore, the 
         stub was in the upd list, while the actual function
         is in this list.  This way, old code will see the stub,
         and future linked code will see the actual function.
     - the remainder of the ones to lookup
         - these are the normal externs (i.e. not in the old
           implementation that we may be updating).

     When we are not dynamically updating a file, then the first
     and third set will be empty (except for UnresolvedSymbol).
*)

open Popsyntax
module Id = Identifier
module PT = Poptype 

(*--------------------------------------------------------------------*)
(* flags *)

let tuple_rep = ref true
let use_unresolved_sym_exn = ref false
let translate_fn_ptrs = ref false
type local_var_convert_type = NONE | SIMPLE | MD5
let rename_local_vars = ref NONE


(*--------------------------------------------------------------------*)
(* helper functions *)

let bogus_loc = (Gcdfec.seg_of_abs 0 0)

let make_exp e = 
  { raw_exp = e; exp_typ = None; exp_loc = bogus_loc;
    exp_un_after = mt_varset; }

let make_stmt rs = 
  { raw_stmt = rs; stmt_loc = bogus_loc;
    un_after = mt_varset; un_before = mt_varset }

let make_termrep t = make_exp(TypInst (make_exp(RepTerm),[t]))

let indirect cap t = TupleType (cap,[t])

let indirect_global cap t =
  match t with
    FnType _ -> t 
  | _ -> indirect cap t

let indirect_term e = make_exp(Primop(AddrOf,[e]))

let is_fn_type t = 
  match t with
    FnType _ -> true
  | _ -> false

let terr loc s =
  let e = Poperr.Etypecheck (Poperr.TypeError s) in
  let e = Gcdfec.mk_err_elab loc (Poperr.error_message e) in
  Gcdfec.post_error e

let new_id =
  let count = ref 0 in 
  (function x -> let c = !count in incr count; x^"__"^(string_of_int c))

let gen_argvars ts =
  let rec aux i ts =
    match ts with
      [] -> []
    | h::t -> (Char.escaped (Char.chr i),h)::(aux (i+1) t) in
  aux (Char.code 'a') ts

(* convert a long name into a single variable name for being in the GOT.
   Changes each ? to a __ *)
let var2field v =
  let len = String.length v in
  let i = ref 0 in
  let questions = 
    let count = ref 0 in
    while !i < len do
      if v.[!i] = '?' then incr count;
      incr i;
    done;
    !count
  in
  let w = String.create (len+questions) in
  let j = ref 0 in
  i:=0;
  while !i < len do
    if v.[!i]='?' 
    then begin w.[!j] <- '_'; w.[!j+1] <- '_'; incr j; end
    else w.[!j] <- v.[!i];
    incr j;
    incr i;
  done;
  w

(* combine a list of statements into a single Seq *)
let combine_stmts sl body =
  List.fold_right (fun s res -> make_stmt(Seq(s,res))) sl body

(* combines a list of decl statements (where the body of the decl is
   assumed to be Skip), and nests them into a single statement
   where the innermost decl has the given body. *)
let combine_decls dl body =
  let rec aux dl =
    match dl with
      [] -> body
    | (s::tl) ->
	(match s.raw_stmt with
	  Decl(v,t,eor,s) ->
	    if s.raw_stmt != Skip then
	      failwith "Decl stmt part is not a Skip"
	    else
	      let s' = aux tl in { s with raw_stmt = Decl(v,t,eor,s') }
	| _ -> failwith "not a Decl") in
  aux dl

(* creates a large && expression arranging the given expressions arguments
   leftwise *)
let and_exp es =
  let rec aux andexp es =
    match es with
      e::t -> 
	aux (make_exp(Conditional (andexp,e,make_exp(Const(Bool false))))) t
    | [] -> andexp in
  match es with
    [e] -> e
  | e1::e2::t -> 
      aux (make_exp(Conditional (e1,e2,make_exp(Const(Bool false))))) t
  | [] -> failwith "popdyntrans: [] not allowed for and_exp"

let add_to_set s sl =
  if not (List.mem s !sl) then
    sl := s::(!sl)
  else
    ()

(* debugging *)
let list_to_string l toString =
  let rec aux = function
      [] -> ""
    | [x] -> toString x
    | (h::t) ->
        let s = toString h in
        (Printf.sprintf "%s, " s) ^ (aux t) in
  aux l

(* ENVIRONMENTS *)

type trans_scope =
    Func of string
  | Def of string
  | Other

type action = 
    LookupEarly | LookupLate | Update | Add

type var_info =
    { actual_varname: string;
      dyn_varname: string;
      var_type: typ;
      dyn_action: action;
      updateable: bool }

type env = { local_env: string list;             (* needed to see if globals
					            are shadowed *)
	     curr_def: trans_scope;              (* def being processed *)
	     externs:                            (* list of externs *)
	       (string, var_info) Hashtbl.t;
	     globals: PT.global_env;             (* global env generated
					            during typechecking *)
             decls: (stmt list) ref;             (* extra local declarations 
						    needed for accessing 
						    exncons through the GOT *)
	     topdecls: (top_decl list) ref;      (* extra toplevel decls
						    needed by inits *)
	     inits: (stmt list) ref;             (* extra statements needed
						    to initialize toplevel
						    variables *)
	     exports:                            (* items to be exported *)
	       (string, var_info) Hashtbl.t;
	     exports_in_GOT: (string list) ref;  (* exports to add to GOT *)
	     export_deps:
	       (string, string list) Hashtbl.t;  (* maps the name of an
						    exported function to
						    the imported vals it
						    requires *)
	     has_init: bool ref;                 (* there is a user-defined
						    init function *)
	     notice_updates: bool;		 (* true if this file is
						    updateable *)
	     static_exports: bool;               (* true if exports should
						    not be made static *)
	     export_local_vars: bool;            (* true if we should
						    register local vars *)
	     export_add_varmap:                  (* export add item names *)
	       (string * string) list;
	     export_upd_varmap:                  (* export upd item names *)
	       (string * string) list;
	     mod_name: string;                   (* module's name *)
	     file_name: string;                  (* module's filename *)
	   }

let pr_varinfo_tab tab =
  let action2string a =
    match a with
      LookupEarly -> "L-early"
    | LookupLate -> "L-late"
    | Update -> "U"
    | Add -> "A" in
  Hashtbl.iter 
    (fun _ 
	{ actual_varname = v;
	  dyn_varname = v';
	  var_type = t;
	  dyn_action = a;
	  updateable = upd; } ->
      Printf.eprintf "(%s:%s) -> %s | %s, upd=%b\n"
	v (typ2string t) v' (action2string a) upd)
    tab

(* indicates the binding of a local variable shadows a global one *)
let bind env v =
  { env with local_env = v::env.local_env }

(* returns true if we need to add an extra indirection *)
let do_indirect env (v,t) =
  match t with
    FnType _ -> 
      (* if it's a non-global, it must be a function pointer *)
      if List.mem v env.local_env then true
      else (* otherwise, check the global environment *)
	not (Popcompenv.is_fun_decl (Popcompenv.env_empty env.globals) v)
  | ExnconType _ -> 
      if Dict.member env.globals.PT.exceptions v then false else true
  | _ -> true

(* The next few functions traverse over a type expression and add an
   extra level of indirection to all function types *)

let rec indirect_fn_typ may_upd do_expopt env t =
  let indirect_fn_typ' = indirect_fn_typ may_upd do_expopt env in
  let indirect_fn_typs' = indirect_fn_typs may_upd do_expopt env in
  let indirect_fn_typopt' = indirect_fn_typopt may_upd do_expopt env in
  match t with
    Evar (vc,tor) -> Evar (vc,ref (indirect_fn_typopt' !tor))
  | ArrayType (t,eopt) -> 
      let t' = indirect_fn_typ' t in
      ArrayType (t', do_expopt env eopt)
  | FnType (vs,t,tl) -> 
      let t' = FnType (vs,indirect_fn_typ' t,indirect_fn_typs' tl) in
      if may_upd && !translate_fn_ptrs then indirect ReadOnly t' else t'
  | TupleType (c,tl) -> TupleType (c,indirect_fn_typs' tl)
  | NamedType (nr, tl) -> NamedType (nr, indirect_fn_typs' tl)
  | MutableTyp tr -> MutableTyp (ref (indirect_fn_typ' !tr))
  | UnresolvedTyId (n,tl) -> UnresolvedTyId (n, indirect_fn_typs' tl)
  | RepType t -> RepType (indirect_fn_typ' t)
  | ExnconType t -> ExnconType (indirect_fn_typ' t)
  | _ -> t

and indirect_fn_typs may_upd do_expopt env ts =
  List.map (indirect_fn_typ may_upd do_expopt env) ts
and indirect_fn_typopt may_upd do_expopt env topt =
  match topt with 
    None -> None 
  | Some t -> Some (indirect_fn_typ may_upd do_expopt env t)
and indirect_fn_typsopt may_upd do_expopt env tlo =
  match tlo with 
    None -> None 
  | Some ts -> Some (indirect_fn_typs may_upd do_expopt env ts)

(* if is an external, returns its type, and whether it's a variable
   (otherwise it's a declaration) *)
let lookup_external env v =
  if List.mem v env.local_env then
    raise Not_found (* bound locally *)
  else
    let { var_type = t; updateable = upd } = 
      Hashtbl.find env.externs v in
    (t,do_indirect env (v,t),upd)

(* same as lookup_external, but for exports *)
let lookup_export env v =
  if List.mem v env.local_env then
    raise Not_found (* bound locally *)
  else
    let { var_type = t } = Hashtbl.find env.exports v in
    (t,do_indirect env (v,t))

let in_varmap v varmap = List.mem_assoc v varmap
let in_varmap_range v varmap =
  List.fold_left 
    (fun res (_,x) -> res || (x = v))
    false varmap

let in_varmaps f v vml =
  List.fold_left
    (fun res varmap -> res || (f v varmap))
    false vml

(* keep track of the values defined by this file; used to generate
   update statements to the dynamic symbol table within init *)
let note_export env s v t =
  if (s = Public) || env.export_local_vars then
    (* first figure out the dynamic symbol name *)
    (let v' =
      try
	List.assoc v env.export_upd_varmap
      with Not_found ->
	(try
	  List.assoc v env.export_add_varmap
	with Not_found ->
	  if s = Static then
	    (match !rename_local_vars with
	      NONE -> v
	    | SIMPLE -> env.mod_name ^ "?Local?" ^ v 
	    | MD5 -> (Digest.file env.file_name) ^ v)
	  else v) in
    Hashtbl.add env.exports v
      { actual_varname = v;
	dyn_varname = v';
	var_type = t;
	dyn_action = 
	  if in_varmap v env.export_add_varmap then Add else Update;
        updateable = env.notice_updates })
  else ()

(* provides the lookup name for any local variables to be looked up
   dynamically *)
let rename_extern_local_var v =
  let get_prefix n =
    let nLen = String.length n in
    let rec aux i =
      if i=nLen then None
      else
	if n.[i] = '?' then
	  Some(String.sub n 0 i,String.sub n (i+1) (nLen - i - 1))
	else aux (i+1) in
    aux 0 in
  let has_prefix n p = 
    let pLen = String.length p in
    if String.length n <= pLen then None
    else 
      let rec aux i = 
	if i=pLen then Some (String.sub n (i+1) (String.length n - pLen - 1))
	else if n.[i] <> p.[i] then None else aux (i+1)
      in
      aux 0 in
  match (has_prefix v "Local") with
    Some v' ->
      (match (get_prefix v') with
	Some (p,v'') ->
	  (match !rename_local_vars with
	    NONE -> v''
	  | SIMPLE -> p ^ "?Local?" ^ v''
	  | MD5 -> (Digest.file (p^".pop")) ^ v'')
      | None -> failwith 
	    (Printf.sprintf "Local variable but no filename: %s" v'))
  | None -> v

(*--------------------------------------------------------------------*)
(*                            the GOT                                 *)
(*--------------------------------------------------------------------*)

let got_name = new_id "GOT"

(* Find all of the external values used by the file and store them
   in the provided hashtable.  Returns the declarations that don't
   refer to external values/types *)

let externs decls may_upd may_upd_v tab =
  let indirect_fn_typ' t = indirect_fn_typ may_upd (fun _ e -> e) () t in
  let indirect_fn_typs' ts = indirect_fn_typs may_upd (fun _ e -> e) () ts in
  let create_varinfo v t =
    let dyn_v = rename_extern_local_var v in
    let dyn_a = 
      if may_upd then
        if may_upd_v v then LookupLate else LookupEarly 
      else LookupLate in
    { actual_varname = v;
      dyn_varname = dyn_v;
      var_type = t;
      dyn_action = dyn_a;
      updateable = may_upd_v v } in
  let rec aux decls ds =
    match ds with
      [] -> decls
    | (ExternVal (v,t),_) :: tl ->
	let t =
	  (match t with 
	    FnType (vs,t,ts) -> FnType (vs,indirect_fn_typ' t,
					indirect_fn_typs' ts)
	  | _ -> t) in
	let t' = indirect_global ReadWrite t in
	Hashtbl.add tab v (create_varinfo v t');
	aux decls tl
    | ((ExceptionDecl (v,s,t),_) as d)::tl ->
	(match s with
	  Extern -> 
	    let t = indirect_fn_typ' t in
	    Hashtbl.add tab v (create_varinfo v (ExnconType t));
	    aux decls tl
	| _ -> aux (d::decls) tl)
    | d::tl -> aux (d::decls) tl in
  let decls = aux [] decls in
  List.rev decls

(* Creates the structure declaration for the GOT.  If this file is
   updateable, we must add an extra indirection for each extern. *)

(* externs:(var * typ) list -> structdecl *)
let got_decl may_upd_v (xtab:(string, var_info) Hashtbl.t) =
  let fields = 
    let fs = ref [] in
    Hashtbl.iter (fun _ { actual_varname = v;
			  var_type = t } -> 
      fs := (v,ReadWrite,
	     if (may_upd_v v) then indirect ReadOnly t else t)::(!fs))
      xtab;
    !fs in
  let type_name = new_id "GOT_t" in
  { st_scope = Static;
    st_name = type_name;
    st_tyvars = [];
    st_possibly_null = false;
    st_fields = fields }

(* default values for GOT *)

(* Thrown in default functions *)
exception AddDefExn
let have_unresolved_sym_excon_decl = ref false
let unresolved_sym_exncon = "UnresolvedSymbol"
let unresolved_sym_excon_decl = 
  (ExceptionDecl(unresolved_sym_exncon, Extern, StringType))

(* for hash-consing default exncon values *)
let def_exncons = ref (Dict.empty compare)

(* Returns an expression, and an optional list of declarations.  If we
   need to use the UnresolvedSymbol exception and it has not yet been
   declared, throw an exception and restart *)

let rec default_initializer sym got may_upd_v global_env t loc un =
  let def_init t = default_initializer sym got may_upd_v global_env t loc un in
  let err() = terr loc ("declaration of type "^(typ2string t)
			^" requires initializer") in
  let abort () = err (); raise Gcdfec.Exit in
  match compress t with
    IntType(true,_) -> make_exp(Const (Int Numtypes.i32_0)), []
  | IntType(false,_) -> 
      make_exp(Cast(t,make_exp(Const (Int Numtypes.i32_0)))), []
  | BooleanType     -> make_exp(Const (Bool false)), []
  | StringType      -> make_exp(Const (String "")), []
  | CharType        -> make_exp(Const (Char '\000')), []
  | FloatType       -> make_exp(Const (Float Numtypes.zero_f32)), []
  | DoubleType      -> make_exp(Const (Double Numtypes.zero_f64)), []
  | ArrayType(t,None) ->
      make_exp(ConstArray([],Some t)), []
  | ArrayType(t,Some e) ->
      failwith "no default init for sized arrays"
(*
   XXX will this even work?  Can't have function calls for globals
        make_exp(FunCall(make_exp(Var "new_array"),
                         ref None, [e;def_init t]))
*)
  | TupleType (c,ts)    -> 
      let es,ds = List.split (List.map def_init ts) in
      let e = make_exp(NewTuple es) in
      let e = 
	if c = ReadOnly then 
	  make_exp(Cast(TupleType(ReadOnly,ts),e))
	else e in
      e, (List.flatten ds)

  | NamedType (n,ts) ->
      (* is n (which should be fully named) nullable *)
      let possibly_null n =
	(try (Dict.lookup global_env.PT.structs n).st_possibly_null with
	  Dict.Absent ->
	    (try snd (Dict.lookup global_env.PT.abstracts n) with
	      Dict.Absent -> false)) in

      (* we can assume n is already complete, since we typechecked already *)
      let name = !n in
      if (possibly_null name) then 
	make_exp(Const(Null)), []
      else 
	begin try 
	  (* First see if it's a structure *)
	  let sd = Dict.lookup global_env.PT.structs name in
	  let init = List.combine sd.st_tyvars ts in
	  let proc_field (f,c,t) =
	    let e,d = def_init (PT.type_subst init t) in
	    d, (None, e)  in
	  let ds, es = List.split (List.map proc_field sd.st_fields) in
	  make_exp (NewStruct (name,ref (Some ts),es)), (List.flatten ds)
	with Dict.Absent ->
	  (* Now check if it's a union *)
	  begin try 
	    let ud = Dict.lookup global_env.PT.unions name in
	    let (field,e_opt), d = 
	      match ud.un_fields with 
		[] -> terr loc "Union with no cases."; raise Gcdfec.Exit 
	      | (f,t)::tl -> 
		  if t=VoidType then (f,None), []
		  else 	      
		    let init = List.combine ud.un_tyvars ts in
		    let e,d = def_init (PT.type_subst init t) in
		    (f,Some e), d
	    in
	    make_exp (NewUnion (name,ref (Some ts),field,e_opt)), d
	  with Dict.Absent -> 
	    (* Otherwise an abstract type -- no good *)
	    terr loc "Abstract types require initializers.";
	    abort ()
	  end
	end
  | FnType (vs, t, ts) ->
      let exn_exp, exn_decls = 
	if !use_unresolved_sym_exn then
	  if not !have_unresolved_sym_excon_decl then
	    (def_exncons := (Dict.empty compare);
	     raise AddDefExn)
	  else
	    (make_exp
	      (if (may_upd_v got) then
		(TupleMember(make_exp
			       (StructMember
				  (make_exp(Var got),
				   unresolved_sym_exncon)),1))
	      else
		(StructMember(make_exp(Var got),
			      unresolved_sym_exncon))),
	     [])
	else 
	  def_init (ExnconType StringType)
      in
      
      let name = new_id "fn" in
      let exnname = new_id "exn" in
      let fndecl = 
	{ fn_static = false;
	  fn_name = name;
	  fn_tyvars = vs;
	  fn_ret_type = t;
	  fn_args = gen_argvars ts;
	  fn_body = make_stmt 
	    (Decl
	       (exnname,ExnconType StringType,
		ref (Some(exn_exp)),
		make_stmt
		  (Exp(make_exp
		     (Raise(make_exp
			(NewExn(exnname,
			   Some (make_exp
			     (Const(String sym))))))))))) } in
      (make_exp(Var name), [(FunDecl fndecl,bogus_loc)]@exn_decls)
  | ExnconType t ->
      let (name,decls) = 
	if not (Dict.member !def_exncons t) then
	  (let n = new_id "exncon" in
	  def_exncons := Dict.insert !def_exncons t n; 
	  (n, [(ExceptionDecl(n,Static,t),bogus_loc)]))
	else
	  let n = Dict.lookup !def_exncons t in
	  (n,[]) in
      (make_exp(Var name), decls)
  | ExnType ->
      let exncon_name, exncon_decl = 
	(try
	  (Dict.lookup !def_exncons VoidType, [])
	with Dict.Absent ->
	  (let name = new_id "exncon" in
	  def_exncons := Dict.insert !def_exncons t name;
	  (name, [(ExceptionDecl(name,Static,VoidType),bogus_loc)]))) in
      
      (make_exp(NewExn (exncon_name,None)),exncon_decl)
  | _ -> abort ()

(* If this is a statically linked file that is nonetheless using a
   GOT, then we make the default values in the GOT point directly to
   the externs. *)

let extern_as_default v t may_upd =
  let e = make_exp(Primop(AddrOf,[make_exp(Var v)])) in
  if may_upd then
    make_exp(NewTuple [e]),[]
  else
    e, []

(* Given the externs for the file and its global environment, returns back 
   the needed declarations for the GOT, its type, and initialization.  
   If the static_exports flag is set, then the initial values of the
   GOT come from the external definitions rather than the default values. *)

(* (var * typ) list * global_env -> topdecl list *)
let rec got (xtab:(string, var_info) Hashtbl.t) may_upd upd_varmap static_exports global_env =
  (* need to clear from previously processed file *)
  def_exncons := (Dict.empty compare);
  
  let may_upd_v v =
    may_upd && (not (in_varmap_range v upd_varmap)) in
  let field_values ts =
    let decls = ref [] in
    let rec aux ts =
      match ts with
	[] -> []
      | (v,_,t)::tl ->
	  let (e,ds) = 
	    if not static_exports then
	      default_initializer 
		v got_name may_upd_v global_env t bogus_loc mt_varset 
	    else 
	      extern_as_default v t (may_upd_v v) in
	  decls := (ds @ !decls);
	  (Some (var2field v),e)::(aux tl) in
    let fields = aux ts in
    (fields,!decls) in

  let decl,(fvals,decls) = 
    try
      let decl = got_decl may_upd_v xtab in
      { decl with st_fields = 
	List.map (fun (v,s,t) -> (var2field v,s,t)) decl.st_fields }, 
      (field_values decl.st_fields)
    with AddDefExn ->
      (have_unresolved_sym_excon_decl := true;
       let _ = externs [(unresolved_sym_excon_decl,bogus_loc)] 
	   may_upd may_upd_v xtab in
       let decl = got_decl may_upd_v xtab in
       { decl with st_fields = 
	 List.map (fun (v,s,t) -> (var2field v,s,t)) decl.st_fields }, 
       (field_values decl.st_fields)) in
  
  let init_exp =
    NewStruct(decl.st_name, ref None, fvals) in
  
  (decls,(StructDecl decl,bogus_loc),
   (GlobalDecl(Static, got_name,
	       NamedType ((ref decl.st_name),[]),
	       ref (Some (make_exp init_exp))),bogus_loc))

(* given a GOT type and value declaration, this adds the given
   globals to both *)
let add_globals_to_got may_upd got_sdecl got_decl gs =
  let fields = 
    List.map (fun (v,t) -> (var2field v,ReadWrite,
			    if may_upd then indirect ReadOnly t else t)) gs in
  let field_vals =
    List.map 
      (fun (v,t) -> let e,_ = extern_as_default v t may_upd in 
      (Some (var2field v), e)) gs in
  match got_sdecl, got_decl with
    (StructDecl sd,loc1), ((GlobalDecl(s,v,t,eor),_) as gd) ->
      (match !eor with
	Some { raw_exp = NewStruct (n,tor,fvals) } ->
	  eor := Some(make_exp(NewStruct (n,tor,fvals@field_vals)));
	  (StructDecl { sd with st_fields = sd.st_fields @ fields },loc1), gd
      |	_ -> failwith "add_globals_to_got: no or improper init exp")
  | _ -> failwith "add_globals_to_got: didn't get GOT structdecl"
      

(*--------------------------------------------------------------------*)
(*                       references to the GOT                        *)
(*--------------------------------------------------------------------*)

(* note --- we assume that typechecking has been run, so that the
   UnionSwitch statement returned by the parser has been appropriately
   translated into an ExnSwitch, NewStruct into NewExn, etc. *)

(* used to keep track of which functions depend on which imported 
   symbols.  This information is currently not used. *)

let note_dependence env v =
  if env.static_exports then ()
  else
    match env.curr_def with
      Func f ->
	let old_deps = 
	  try
	    let l = Hashtbl.find env.export_deps f in
	    Hashtbl.remove env.export_deps f;
	    l
	  with Not_found ->
	    [] in
	let new_deps =
	  if List.mem v old_deps then old_deps
	  else v::old_deps in
	Hashtbl.add env.export_deps f new_deps
    | _ -> ()

(* don't make pop_main static if it exists *)

let pop_main = "pop_main"

(* VARIABLES *)

(* redirects a variable to go through the GOT; the check as to whether
   to perform this translation should occur in the caller *)
let dyntrans_var env v t needs_indirect may_upd =
  (* change the expression to go through the GOT *)
  let e = (StructMember(make_exp(Var got_name),(var2field v))) in
  let e = if needs_indirect then TupleMember(make_exp e,1) else e in
  let e = if may_upd then
    if !translate_fn_ptrs && is_fn_type t then e
    else TupleMember(make_exp e,1)
  else e in

  match env.curr_def with
    Func _ -> (* note the dependence on the current function *)
      note_dependence env v; e
  | Def lv ->  (* toplevel initialization statement *)
    (* Since it will now be non-constant, it needs to be
       added to the init function; we'll use a dummy initialization
       function at first, to make the typechecker happy. *)
      let (def_e,tds) = 
	let may_upd_v v = may_upd in
	default_initializer v got_name may_upd_v env.globals 
	  (if may_upd && !translate_fn_ptrs && is_fn_type t then 
	    indirect ReadOnly t
	  else t) bogus_loc mt_varset in
      (* generate the dyninit initialization statement *)
      (* XXX this won't work if the statement is nested---that is
	 *(int (int)) foo = ^(bar)
	 will end up with an initialization statement of
	   foo = GOT.bar
	 when it should really be
	   foo.1 = GOT bar 
         Seems like we need to track the expression somehow
         so that we know how to properly generate this statement. *)
      let init_stmt = 
	make_stmt 
	  (Exp (make_exp(AssignOp(make_exp(Var lv),None,
				  make_exp e)))) in
      (env.inits) := init_stmt :: (!(env.inits));
      (env.topdecls) := tds @ (!(env.topdecls));
      def_e.raw_exp
  | _ -> failwith "in dyntrans_var when not in fun or toplevel init"

(* TYPES *)

let rec dyntrans_typ env t = 
  indirect_fn_typ env.notice_updates dyntrans_expopt env t
and dyntrans_typs env ts = 
  indirect_fn_typs env.notice_updates dyntrans_expopt env ts
and dyntrans_typopt env topt = 
  indirect_fn_typopt env.notice_updates dyntrans_expopt env topt
and dyntrans_typsopt env tsopt = 
  indirect_fn_typsopt env.notice_updates dyntrans_expopt env tsopt

(* EXPRESSIONS *)

and dyntrans_exp env exp =
  let dyntrans_exp' = dyntrans_exp env in
  let dyntrans_exps' = dyntrans_exps env in
  let dyntrans_expopt' = dyntrans_expopt env in
  let dyntrans_stmt' = dyntrans_stmt env in
  let dyntrans_typ' = dyntrans_typ env in
  let dyntrans_typs' = dyntrans_typs env in
  let dyntrans_typopt' = dyntrans_typopt env in
  let dyntrans_typsopt' = dyntrans_typsopt env in
  let pr_typsopt tlo =
    match tlo with
      Some ts -> 
	Printf.eprintf "<";
	List.iter (function t ->
	  let s = typ2string t in
	  Printf.eprintf "%s " s) ts;
	Printf.eprintf ">\n";
    | None -> () in
  let exp_typ exp = 
    (match exp.exp_typ with
      None -> failwith ("can't get type for "^(exp2string exp))
    | Some t -> t) in
  let dyntrans_funcall avoid_indirect fexp tlor args =
    let fexp' = 
      let e = dyntrans_exp' fexp in
      if env.notice_updates && !translate_fn_ptrs && 
	(not avoid_indirect) then
	make_exp(TupleMember(e,1))
      else e in
    FunCall (fexp', ref (dyntrans_typsopt' !tlor), dyntrans_exps' args) in
  let e' =
    match exp.raw_exp with
      Const c -> exp.raw_exp
    | ConstArray (el, topt) -> 
	ConstArray (dyntrans_exps' el, dyntrans_typopt' topt)
    | Var v -> 
	(* see if we need to indirect this reference through the GOT *)
	(try
	  let t,needs_indirect,may_upd = lookup_external env v in
	  dyntrans_var env v t needs_indirect may_upd
	with Not_found ->
	  (* non-external variable *)
	  (if env.notice_updates && !translate_fn_ptrs then
	    try
	      let t,needs_indirect = lookup_export env v in
	      if (not (do_indirect env (v,t))) then
		(add_to_set v env.exports_in_GOT;
		 dyntrans_var env v t needs_indirect env.notice_updates)
	      else
		Var v
	    with Not_found ->
	      (* must be a local *)
	      (* let _ = Printf.eprintf "dyntrans_var: local var %s\n" v in *)
	      let t = exp_typ exp in
	      if (not (do_indirect env (v,t))) then
		let t = dyntrans_typ' t in
		Cast (t, make_exp(NewTuple [make_exp (Var v)]))
	      else
		Var v
	  else
	    Var v))
    | Dlsym -> 
	failwith "popdyntrans: dlsym not supported in loadable file"
    | Primop (op, el) -> Primop (op, dyntrans_exps' el)
    | Conditional (be,ife,elsee) -> 
	Conditional (dyntrans_exp' be, dyntrans_exp' ife, dyntrans_exp' elsee)
    | AssignOp (lhs,opopt,rhs) ->
	AssignOp (dyntrans_exp' lhs,opopt,dyntrans_exp' rhs)

    (* deal with function expression cases specially *)
    (* XXX add cases for codegen, anon funs, ... *)
    | FunCall ({ raw_exp = Var v } as fexp, tlor, args) ->
	(try
	  (* need to indirect this reference through the GOT *)
	  let t,needs_indirect,may_upd = lookup_external env v in
	  dyntrans_funcall (not may_upd) fexp tlor args
	with Not_found ->
	  (* non-external variable *)
	  if do_indirect env (v,exp_typ fexp) then
	    dyntrans_funcall false fexp tlor args
	  else
	    FunCall (fexp, ref (dyntrans_typsopt' !tlor), 
		     dyntrans_exps' args))
    | FunCall (fexp, tlor, args) -> dyntrans_funcall false fexp tlor args
    | TypInst (e,ts) -> TypInst (dyntrans_exp' e,dyntrans_typs' ts)
    | NewStruct (n, tlor ,fnoptelist) ->
	let dyntrans_field (fopt, e) = (fopt, dyntrans_exp' e) in
	NewStruct (n, ref (dyntrans_typsopt' !tlor), 
		   List.map dyntrans_field fnoptelist)
    | StructMember (e',n) -> StructMember (dyntrans_exp' e',n)
    | NewUnion (n, tlor, nf, eopt) ->
	NewUnion (n, ref (dyntrans_typsopt' !tlor), nf, dyntrans_expopt' eopt)
    | UnionMember (e',n) -> UnionMember (dyntrans_exp' e',n)
    | NewTuple el -> NewTuple (dyntrans_exps' el)
    | TupleMember (e',x) -> TupleMember (dyntrans_exp' e',x)
    | NewAbstype (n, tlor, tlor2, e) ->
	NewAbstype (n, ref (dyntrans_typsopt' !tlor), 
		    ref (dyntrans_typsopt' !tlor2), 
		    dyntrans_exp' e)
    | Subscript (e',ie) -> Subscript (dyntrans_exp' e', dyntrans_exp' ie)
    | Codegen fd -> Codegen (dyntrans_fundecl env fd)
    | Fill e -> Fill (dyntrans_exp' e)
    | NewExn (v,eopt) -> 
	let v' = dyntrans_exncon env v in
	NewExn(v', dyntrans_expopt' eopt)
    | Raise e -> Raise (dyntrans_exp' e)
    | SeqExp el -> SeqExp (dyntrans_exps' el)
    | Nop -> exp.raw_exp
    | Cast (t,e') -> Cast (dyntrans_typ' t,dyntrans_exp' e')
    | Fun fd -> Fun (dyntrans_fundecl env fd)
    | RepTerm -> exp.raw_exp
  in
  { exp with raw_exp = e' }

and dyntrans_exps env el = List.map (dyntrans_exp env) el
and dyntrans_expopt env eopt =
  match eopt with None -> None | Some e -> Some (dyntrans_exp env e)

(* if the exception name is an external, then create a variable declaration
   to hold its indirection through the GOT and return that variable. *)
and dyntrans_exncon env v =
  try
    let t,needs_indirect,may_upd = lookup_external env v in
    (* note the dependence on the current function *)
    note_dependence env v;
    (* need to create variable to hold the reference in the GOT *)
    let e = 
      (make_exp
	 (StructMember
	    (make_exp(Var got_name),(var2field v)))) in
    (* let e = if needs_indirect then make_exp(TupleMember(e,1)) else e in *)
    (* don't need to check varmap here because we know it's not
       a function *)
    let e = if may_upd then make_exp(TupleMember(e,1)) else e in
    let v' = new_id "exnconv" in
    let s = make_stmt(Decl (v', t, ref (Some e), make_stmt(Skip))) in
    env.decls := (s::!(env.decls));
    v'
  with Not_found -> v

(* STATEMENTS *)

and dyntrans_stmt env stmt =
  let dyntrans_exp' = dyntrans_exp env in
  let dyntrans_exps' = dyntrans_exps env in
  let dyntrans_expopt' = dyntrans_expopt env in
  let dyntrans_typ' = dyntrans_typ env in
  let dyntrans_typopt' = dyntrans_typopt env in
  let dyntrans_stmt' = dyntrans_stmt env in
  let dyntrans_stmtopt' = dyntrans_stmtopt env in
  let dyntrans_arm (x,s) = (x,dyntrans_stmt' s) in
  let dyntrans_sarm sa = 
    (* figure out all of the variables bound in the context of the arm *)
    let arm_vars p =
      let do_prim_pat p =
	match p with
	  Var_pat (v,_) -> [v]
	| _ -> [] in
      match p with
      	No_pat -> []
      |	Prim_pat p -> do_prim_pat p
      |	Tuple_pat pl ->
	  List.flatten (List.map do_prim_pat pl) in
    let env' =
      List.fold_left (fun env' v -> bind env' v) env (arm_vars sa.arm_pat) in
    (* figure out if the field name is an external exception *)
    let v' = (* assumes no name clashes with structs, abstypes, etc. *)
      dyntrans_exncon env' sa.arm_field in
    { sa with arm_field = v'; arm_body = dyntrans_stmt env' sa.arm_body } in

  let s' = 
    match stmt.raw_stmt with
      Skip -> stmt.raw_stmt
    | Exp e -> Exp (dyntrans_exp' e)
    | Seq (s,s2) -> Seq (dyntrans_stmt' s, dyntrans_stmt' s2)
    | Return eopt -> Return (dyntrans_expopt' eopt)
    | IfThenElse (e,s1,s2) ->
	IfThenElse (dyntrans_exp' e,dyntrans_stmt' s1,dyntrans_stmt' s2)
    | While (e,s) -> While (dyntrans_exp' e,dyntrans_stmt' s)
    | Break vopt -> stmt.raw_stmt
    | Continue vopt -> stmt.raw_stmt
    | For (e1,e2,e3,s) ->
	For (dyntrans_exp' e1,dyntrans_exp' e2,dyntrans_exp' e3, 
	     dyntrans_stmt' s)
    | IntSwitch (e, isl, s) ->
	IntSwitch (dyntrans_exp' e, List.map dyntrans_arm isl, 
		   dyntrans_stmt' s)
    | CharSwitch (e,csl,s) ->
	CharSwitch (dyntrans_exp' e, List.map dyntrans_arm csl, 
		    dyntrans_stmt' s)
    | UnionSwitch (e,sal,sopt) ->
	UnionSwitch (dyntrans_exp' e, List.map dyntrans_sarm sal, 
		     dyntrans_stmtopt' sopt)
    | ExnSwitch (e,sal,sopt) ->
	ExnSwitch (dyntrans_exp' e,List.map dyntrans_sarm sal,
		   dyntrans_stmtopt' sopt)
    | Decl (v,t,eor,s) -> Decl (v,dyntrans_typ' t,
				ref (dyntrans_expopt' !eor),
				dyntrans_stmt (bind env v) s)
    | Label (v,s) -> Label (v,dyntrans_stmt' s)
    | Do (s,e) -> Do (dyntrans_stmt' s,dyntrans_exp' e)
    | TryHandle (s1,v,s2) -> TryHandle (dyntrans_stmt' s1,v,
					dyntrans_stmt (bind env v) s2)
    | TryCatchFinally (s,sal,defopt,finopt) ->
	TryCatchFinally (dyntrans_stmt' s,List.map dyntrans_sarm sal,
			 dyntrans_stmtopt' defopt,dyntrans_stmtopt' finopt)
    | Cut s -> Cut (dyntrans_stmt' s)
    | Splice s -> Splice (dyntrans_stmt' s)
    | With (v,tor,vs,e,s) -> 
	With (v,ref (dyntrans_typopt' !tor), vs, 
	      dyntrans_exp' e, dyntrans_stmt (bind env v) s)
  in
  { stmt with raw_stmt = s' }
  
and dyntrans_stmtopt env eopt =
  match eopt with None -> None | Some e -> Some (dyntrans_stmt env e)

(* DECLARATIONS *)

and dyntrans_topdecl env ((rawdecl,loc) as decl) =
  let dyntrans_expopt' = dyntrans_expopt env in
  let dyntrans_typ' = dyntrans_typ env in
  let dyntrans_typs' = dyntrans_typs env in
  let d' =
    match rawdecl with
      FunDecl fd -> 
	if fd.fn_name = "init" then (env.has_init) := true;
	FunDecl (dyntrans_fundecl { env with curr_def = Func fd.fn_name } fd)
    | StructDecl sd ->
	StructDecl
	  { sd with st_fields =
	    List.map 
	      (function (fn,c,t) -> (fn,c,dyntrans_typ' t)) sd.st_fields }
    | UnionDecl ud ->
	UnionDecl
	  { ud with un_fields =
	    List.map 
	      (function (fn,t) -> (fn,dyntrans_typ' t)) ud.un_fields }
    | ExceptionDecl (v,s,t) -> 
	let t = dyntrans_typ' t in
	(* note_export env s v (ExnconType t); *)
	let s =
	  if s = Public then
	    if not env.static_exports then Static else s
	  else s in
	ExceptionDecl (v,s,t)
    | ExternVal (v,t) -> 
	if env.notice_updates && !translate_fn_ptrs then
	  let t = 
	    (match t with 
	      FnType (vs,t,ts) -> FnType (vs,dyntrans_typ' t,dyntrans_typs' ts)
	    | _ -> t) in
	  ExternVal (v,t)
	else
	  rawdecl
    | AbsDecl ad -> 
	if env.notice_updates && !translate_fn_ptrs then
	  AbsDecl { ad with abs_defn = dyntrans_typ' ad.abs_defn }
	else
	  rawdecl
    | ExternType (_, _, _) -> rawdecl
    | GlobalDecl (s,v,t,eor) ->
	let t = dyntrans_typ' t in
	(* note_export env s v t; *)
	let s =
	  if s = Public then 
	    if not env.static_exports then Static else s
	  else s in
	GlobalDecl (s,v,t,
		    ref (dyntrans_expopt { env with curr_def = Def v } !eor))
    | PrefixDecl (v,tds) -> 
	PrefixDecl (v, List.map (dyntrans_topdecl env) tds)
    | OpenDecl (v,tds) -> 
	OpenDecl (v, List.map (dyntrans_topdecl env) tds)
  in
  (d',loc)

and dyntrans_fundecl env fd =
  let dyntrans_typ' = dyntrans_typ env in
  let dyntrans_typs' = dyntrans_typs env in

  (* translate the body *)
  let env' =
    List.fold_left (fun env' v -> bind env' v) env (List.map fst fd.fn_args) in
  let body' = (dyntrans_stmt env' fd.fn_body) in
  (* now add in all of the declarations for exn switches *)
  let body'' = combine_decls !(env.decls) body' in
  (env.decls) := []; (* reset for the next function decl *)

  (* translate the declaration (may need to indirect some fn ptrs) *)
  let ret_type = dyntrans_typ' fd.fn_ret_type in
  let args = List.map (function (v,t) -> (v,dyntrans_typ' t)) fd.fn_args in
  let t = FnType (fd.fn_tyvars, ret_type, List.map snd args) in

  (* note the export and return the transformed decl *)
(*
  note_export env (if fd.fn_static then Public else Static) fd.fn_name t;
*)
  let s = 
    if fd.fn_static && fd.fn_name <> pop_main && not env.static_exports then
      false
    else
      fd.fn_static in
  { fd with 
    fn_static = s; fn_ret_type = ret_type; fn_args = args; fn_body = body'' }

(*--------------------------------------------------------------------*)
(*                         the init function                          *)
(*--------------------------------------------------------------------*)

(* gathers all of the exported toplevel variables *)

let note_exports env ((rawdecl,loc) as decl) =
  let do_exp _ e = e in
  match rawdecl with
    FunDecl fd -> 
      let ret_type = 
	indirect_fn_typ env.notice_updates do_exp env fd.fn_ret_type in
      let args = List.map 
	  (function (v,t) -> 
	    (v,indirect_fn_typ env.notice_updates do_exp env t)) fd.fn_args in
      let t = FnType (fd.fn_tyvars, ret_type, List.map snd args) in
      let s = if fd.fn_static then Public else Static in
      note_export env s fd.fn_name t
  | ExceptionDecl (v,s,t) -> 
      let t' = indirect_fn_typ env.notice_updates do_exp env t in
      note_export env s v (ExnconType t')
  | GlobalDecl (s,v,t,eor) ->
      let t' = indirect_fn_typ env.notice_updates do_exp env t in
      note_export env s v t'
  | _ -> ()

(* assumes that the exports portion of the environment has been 
   filled in by the previous stage. *)
    
let gen_init env do_export do_lookups init_id =

  (********************)
  (* helper functions *)
  (********************)

  (* names lookup and update functions *)
  let lookup_id = new_id "lookup" in 
  let update_id = new_id "update" in 
  let add_id = new_id "add" in 
  let upd_flag =  new_id "is_updated" in

  (* name for flag used by mutually recursive version *)
  let vflag_id v = (var2field v)^"_flag" in
  let bool_exp v b =
    make_exp(AssignOp(make_exp(Var v),
		      None,
		      make_exp (Const(Bool b)))) in
  let bool_stmt v b =
    make_stmt(Exp(bool_exp v b)) in

  let gen_flag_decls ds b =
    let rec aux decls ds =
      match ds with
	(v,_)::tl ->
	  let vflag = vflag_id v in
	  let d = 
	    GlobalDecl (Static,vflag,BooleanType,
			ref (Some (make_exp(Const(Bool b))))) in
	  aux ((d,bogus_loc)::decls) tl
      |	[] -> decls
    in
    aux [] ds in

  (* given a list of variables, generates a linking statement
     for each one by calling gen_stm *)
  let gen_linking_stms gen_stm xtab do_gen =
    let stms = ref [] in
    Hashtbl.iter
      (fun _ ({ actual_varname = act_v;
		dyn_varname = dyn_v;
		var_type = t;
	        updateable = upd; } as vi) ->
        if (do_gen vi) then
	  stms := (gen_stm act_v dyn_v t upd)::(!stms)
	else
	  ())
      xtab;
    !stms in
  
  (* Generates a list of statements to lookup each import and fill
     its value in the GOT *)
  let gen_imports xtab do_gen =
    let gen_import fun_v sym_v t may_upd =
      let t = 
	if !tuple_rep || env.notice_updates then indirect ReadWrite t else t in
      let termrep = make_termrep t in
      let lookup_funcall = 
	make_exp(FunCall
		   (make_exp(Var lookup_id),
		    ref None,
		    [ make_exp(Const (String sym_v));
		      termrep ])) in
      let update_got =
	make_exp(AssignOp
		   (make_exp(StructMember
			       (make_exp(Var got_name),
				(var2field fun_v))),
		    None,
		    if !tuple_rep && (not may_upd) then
		       make_exp(TupleMember(lookup_funcall,1))
		    else
		      lookup_funcall)) in
      (make_stmt(Exp update_got)) in
    gen_linking_stms gen_import xtab do_gen in

  (* Generates a list of statements to register each export.  We need
     to add an extra indirection to be able to look up the address
     of a symbol *)

  let gen_exports () =
    let gen_export export_fun_id fun_v sym_v t may_upd =
      try
	let needs_indirect = do_indirect env (fun_v,t) in
	let termrep = 
	  make_termrep 
	    (let t = (if needs_indirect then indirect ReadWrite t else t) in
	    if env.notice_updates || !tuple_rep then 
	      indirect ReadWrite t else t) in
	let funcall = 
	  make_exp(FunCall
		     (make_exp(Var export_fun_id),
		      ref None,
		      [ make_exp(Const (String sym_v));
			termrep;
			if needs_indirect then 
			  indirect_term (make_exp(Var fun_v))
			else 
			  (make_exp(Var fun_v)) ])) in
	make_stmt(Exp funcall)
      with Not_found ->
	failwith ("error -- could not find external "^fun_v) in

    let stms = 
      (gen_linking_stms (gen_export update_id) 
	 env.exports (fun { dyn_action = a } -> a = Update)) @
      (gen_linking_stms (gen_export add_id)
	 env.exports (fun { dyn_action = a } -> a = Add)) in

    if stms = [] then []
    else
      (let upd_stmt = make_stmt(Exp (bool_exp (vflag_id upd_flag) true)) in
      let if_body = combine_stmts stms upd_stmt in
      let if_stmt =
	make_stmt(IfThenElse(make_exp
			       (Primop(Not,
				       [make_exp(Var(vflag_id upd_flag))])),
			     if_body,
			     make_stmt(Skip))) in
      [if_stmt])
  in

  (* Calls the user-defined init function, if it exists *)
  let gen_call_user_init () =
    if !(env.has_init) then
      make_stmt(Exp(make_exp(FunCall
				(make_exp(Var "init"),
				 ref None,
				 []))))
    else
      make_stmt(Skip) in

  (************************)
  (* The dyninit function *)
  (************************)

  (* The makeup of the init function is as follows:
     - initial lookup statements
       These are needed to get old versions of functions about to be
       replaced
     - export statements
       Registers the symbols to be exported by this file
     - lookup statements
       Most of the lookups for imported symbols occur here
     - special initialization statements
       These initialize toplevel variables whose initialization
       statements were constant before the transformation but now
       are not
     - call to user-defined init function (if it exists)
  *)

  (* generate import statements *)
  let (start_import_stms, finish_import_stms) = 
    if do_lookups then 
      (gen_imports env.externs (fun { dyn_action = a } -> a = LookupEarly),
       gen_imports env.externs (fun { dyn_action = a } -> a = LookupLate))
    else
      ([],[]) in

  (* Add globally exported functions to GOT for function pointers *)
  let finish_import_stms =
    if env.notice_updates && !translate_fn_ptrs then
      let export_import_stms =
	let global_env = Popcompenv.env_empty env.globals in
	gen_imports env.exports
 	  (fun { actual_varname = v } -> 
	    List.mem v !(env.exports_in_GOT) &&
	    Popcompenv.is_fun_decl global_env v) in
      export_import_stms @ finish_import_stms
    else finish_import_stms in

  (* generate export statements *)
  let export_stms = if do_export then gen_exports () else [] in

  (* generate call to user-defined init function *)
  let init_stm = gen_call_user_init () in

  (* generate the declaration for the dyninit function *)
  let fd =
    { fn_static = true;
      fn_name = init_id;
      fn_tyvars = [];
      fn_ret_type = VoidType;
      fn_args = [ (lookup_id, FnType(["a"],
				     (if env.notice_updates || !tuple_rep then
				       TupleType (ReadWrite,[VarType "a"])
				     else 
				       VarType "a"),
				     [StringType;
				       RepType
					 (if env.notice_updates 
					  || !tuple_rep then
					   TupleType (ReadWrite,[VarType "a"])
					 else 
					   VarType "a")]));
		  (add_id, FnType(["a"],
				  VoidType,
				  [ StringType;
				    RepType
				      (if env.notice_updates 
				     || !tuple_rep then
					TupleType (ReadWrite,[VarType "a"])
				      else 
					VarType "a");
				    VarType "a"]));
		  (update_id, FnType(["a"],
				     VoidType,
				     [ StringType;
				       RepType
					 (if env.notice_updates 
					  || !tuple_rep then
					   TupleType (ReadWrite,[VarType "a"])
					 else 
					   VarType "a");
				       VarType "a"])) ];
      fn_body = (combine_stmts 
		   (start_import_stms@export_stms@
		    finish_import_stms@ (!(env.inits))) 
		   init_stm);
    }
  in
  let init_decl = FunDecl fd,bogus_loc in
  if export_stms != [] then
    let update_flag_decl = gen_flag_decls [(upd_flag,())] false in
    update_flag_decl@[init_decl]
  else
    [init_decl]
  
(*--------------------------------------------------------------------*)
(* Takes as arguments the list of declarations that make up the program.
   Performs typechecking, and then returns a new decl list (which
   must be typechecked by the caller) representing the dynamically
   loadable code. *)

type dyntrans_flags =
    { do_import: bool;
      do_export: bool;
      do_notice_updates: bool;
      do_static_exports: bool;
      do_export_local_vars: bool;
      add_varmap: (string * string) list;
      upd_varmap: (string * string) list;
    }

let dyntrans ifc_only decls flags filename modname =
  Printf.printf "Transforming file: %s\n" modname; flush stdout;
  let init_id = "dyninit_"^modname in
  (* first type check *)
  let (decls, global_env) = Poptype.type_check ifc_only decls in
  let env, decls =
    if flags.do_import then
      (* filter out the externs *)
      (let xtab = Hashtbl.create 59 in
       let decls' = 
	 let may_upd_v v =
	   flags.do_notice_updates && 
	   (not (in_varmap_range v flags.upd_varmap)) in
  	 externs decls flags.do_notice_updates may_upd_v xtab in

      (* add the GOT stuff *)
      let (init_decls, got_sdecl, got_decl) =
	got
	  xtab
	  flags.do_notice_updates 
	  flags.upd_varmap
	  flags.do_static_exports 
	  global_env in
(*
      Printf.eprintf "externs are:\n"; pr_varinfo_tab xtab;
*)
      (* change references to be through the GOT *)
      let env = { local_env = [];
		  curr_def = Other;
		  externs = xtab;
		  globals = global_env;
		  decls = ref [];
		  inits = ref [];
		  topdecls = ref [];
		  exports = Hashtbl.create 59;
		  exports_in_GOT = ref [];
		  export_deps = Hashtbl.create 59;
		  has_init = ref false;
		  notice_updates = flags.do_notice_updates;
		  static_exports = flags.do_static_exports;
		  export_local_vars = flags.do_export_local_vars;
		  export_add_varmap = flags.add_varmap;
		  export_upd_varmap = flags.upd_varmap;
		  mod_name = modname;
		  file_name = filename } in
      (* note all of the exports before doing the translation *)
      List.iter (fun d -> note_exports env d; ()) decls;

      (* now do the translation *)
      let decls = 
	List.map (dyntrans_topdecl env) 
	  (if flags.do_static_exports then decls else decls') in
      
      (* add global function vars to GOT *)
      let got_sdecl, got_decl = 
	if (env.notice_updates && !translate_fn_ptrs) then
	  let gs =
	    let global_env = Popcompenv.env_empty env.globals in
	    let s = ref [] in
	    Hashtbl.iter (fun _ { actual_varname = v; var_type = t } ->
	      if List.mem v !(env.exports_in_GOT) &&
		 (Popcompenv.is_fun_decl global_env v) then
		s := (v,t)::(!s)
	      else ()) env.exports;
	    !s in
	  add_globals_to_got env.notice_updates got_sdecl got_decl gs
	else
	  got_sdecl, got_decl in

      let got_decls = init_decls @ [ got_sdecl; got_decl ] in
      env, got_decls@decls@(!(env.topdecls)))
    else
      (let env = { local_env = [];
		   curr_def = Other;                  (* not used *)
		   externs = Hashtbl.create 11;       (* not used *)
		   globals = global_env;
		   decls = ref [];                    (* not used *)
		   inits = ref [];                    (* not used *)
		   topdecls = ref [];                 (* not used *)
		   exports = Hashtbl.create 59;
		   exports_in_GOT = ref [];           (* not used *)
		   export_deps = Hashtbl.create 11;   (* not used *)
		   has_init = ref false;
		   notice_updates = flags.do_notice_updates;
		   static_exports = flags.do_static_exports;
		   export_local_vars = flags.do_export_local_vars;
		   export_add_varmap = flags.add_varmap;
		   export_upd_varmap = flags.upd_varmap;
		   mod_name = modname;
		   file_name = filename } in
      (* generate export list only *)
      List.iter (fun d -> note_exports env d; ()) decls;
      env, decls) in

  (* add initialization function *)
  let initdecls = 
    gen_init 
      env flags.do_export 
      (* even if we have static exports (and thus use the static
	 linker to fill in the values in the table), we want to
	 do the lookups if we notice updates, so that we get the
	 tuple stored in the symbol table *)
      ((not flags.do_static_exports) || flags.do_notice_updates)
      init_id 
  in
  decls@initdecls
